﻿using System;
using System.Security;
using Microsoft.VisualStudio.TestTools.UnitTesting;

namespace Vista.RPC
{
    /// <summary>
    /// This is a base helper class that can be inherited by actual unit tests.
    /// Examples that inherit from this class are located in the Procedures folder.
    /// </summary>
    public class BaseRpcTest : TestConfig
    {
        /// <summary>
        /// Connect to the TEST VISTA using the A/V code pair set in the App.config and return the rpc client
        /// so the parent class can test
        /// </summary>
        /// <returns>A connected and logged in rpc client</returns>
        public static RpcClient ConnectAndLogin(bool useTestClaims = false)
        {
            SecureString accessCode = null;
            SecureString verifyCode = null;
            RpcClient rpcClient = null;

            var logger = RpcLogging.CreateLogger<RpcClient>();

            if (useTestClaims)
            {
                accessCode = Get(Config.TEST_CLAIMS_ACCESS_CODE).ToSecureString();
                verifyCode = Get(Config.TEST_CLAIMS_VERIFY_CODE).ToSecureString();

                rpcClient = new RpcClient(logger, Get(Config.TEST_CLAIMS_HOSTNAME), Get(Config.TEST_CLAIMS_PORT));
            }
            else
            {
                accessCode = Get(Config.TEST_VISTA_ACCESS_CODE).ToSecureString();
                verifyCode = Get(Config.TEST_VISTA_VERIFY_CODE).ToSecureString();

                rpcClient = new RpcClient(logger, Get(Config.TEST_VISTA_HOSTNAME), Get(Config.TEST_VISTA_PORT));
            }

            try
            {
                rpcClient.Connect();
                rpcClient.Login(accessCode, verifyCode);
            }
            catch (Exception e)
            {
                Console.WriteLine(e.StackTrace);
                Assert.Fail(e.Message);
            }            

            return rpcClient;
        }

        /// <summary>
        /// Gets an instance of the rpc client that is set to use the TEST VISTA instance.
        /// </summary>
        /// <returns>An unconnected rpc client</returns>
        public static RpcClient GetRpcClient(bool useTestClaims = false)
        {
            RpcClient rpcClient = null;

            var logger = RpcLogging.CreateLogger<RpcClient>();

            if (useTestClaims)
            {
                rpcClient = new RpcClient(logger, Get(Config.TEST_CLAIMS_HOSTNAME), Get(Config.TEST_CLAIMS_PORT));
            }
            else
            {
                rpcClient = new RpcClient(logger, Get(Config.TEST_VISTA_HOSTNAME), Get(Config.TEST_VISTA_PORT));
            }
            
            return rpcClient;
        }

        /// <summary>
        /// Gets an instance of the rpc client that has connected to the TEST VISTA instance but has failed to login.
        /// </summary>
        /// <returns>A connected but unauthenticated rpc client</returns>
        public static RpcClient ConnectAndBadLogin()
        {
            var logger = RpcLogging.CreateLogger<RpcClient>();
            RpcClient rpcClient = new RpcClient(logger, Get(Config.TEST_VISTA_HOSTNAME), Get(Config.TEST_VISTA_PORT));
            rpcClient.Connect();
            rpcClient.Login("BAD".ToSecureString(), "AV".ToSecureString());

            return rpcClient;
        }
    }
}
