﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Vista.RPC.Parameters;
using Vista.RPC.Exceptions;

namespace Vista.RPC.Procedures.DDR
{
    public class DdrLister : RemoteProcedure
    {
        /// <summary>
        /// 
        /// </summary>
        /// <param name="file"></param>
        /// <param name="iens"></param>
        /// <param name="fields"></param>
        /// <param name="flags"></param>
        /// <param name="number"></param>
        /// <param name="from"></param>
        /// <param name="part"></param>
        /// <param name="index"></param>
        /// <param name="screen"></param>
        /// <param name="identifier"></param>
        public DdrLister(string file, string iens = "", string fields = "", string flags = "", string number = "", 
            string from = "", string part = "", string index = "", string screen = "", 
            string identifier = "") : base("DDR LISTER")
        {
            arguments = new Dictionary<string, string>();

            arguments.Add("\"FILE\"", file);

            if (!String.IsNullOrEmpty(iens))
                arguments.Add("\"IENS\"", iens);
            if (!String.IsNullOrEmpty(fields))
                arguments.Add("\"FIELDS\"", fields);
            if (!String.IsNullOrEmpty(flags))
            {
                arguments.Add("\"FLAGS\"", flags.ToUpper());
                Flags = flags.ToUpper();
            }
            if (!String.IsNullOrEmpty(number))
                arguments.Add("\"MAX\"", number);
            if (!String.IsNullOrEmpty(from))
                arguments.Add("\"FROM\"", from);
            if (!String.IsNullOrEmpty(part))
                arguments.Add("\"PART\"", part);
            if (!String.IsNullOrEmpty(index))
                arguments.Add("\"XREF\"", index);
            if (!String.IsNullOrEmpty(screen))
                arguments.Add("\"SCREEN\"", screen);
            if (!String.IsNullOrEmpty(identifier))
                arguments.Add("\"ID\"", identifier);

            ListParameter listParameter = new ListParameter(arguments);
            AddParameter(listParameter);
        }

        public string Flags
        {
            get;
            private set;
        }

        private Dictionary<string, string> arguments;

        public override void AfterCall()
        {
            if (Response.StartsWith("[BEGIN_diERRORS]"))
            {
                parseErrors();
            }
            else if (!String.IsNullOrEmpty(Flags) && Flags.ToUpper().Contains("P"))
            {
                parsePackedResponse();
            }
            else
            {
                parseResponse();
            }
        }

        private Dictionary<string, Dictionary<string, string>> listerResult = new Dictionary<string, Dictionary<string, string>>();
        private Dictionary<int, string> listerMap = new Dictionary<int, string>();

        private void parsePackedResponse()
        {
            string[] results = Response.Split("\r\n");
            bool mapIsNext = false;
            bool dataIsNext = false;

            foreach (string line in results)
            {
                if (line == "[MAP]")
                {
                    mapIsNext = true;
                    continue;
                }
                else if (mapIsNext)
                {
                    mapIsNext = false;

                    string[] piecedMap = line.Split("^");
                    for (int i = 0; i < piecedMap.Length; i++)
                    {
                        listerMap.Add(i, piecedMap[i]);
                    }

                    continue;
                }

                if (line == "[BEGIN_diDATA]")
                {
                    dataIsNext = true;
                    continue;
                }
                else if (line == "[END_diDATA]")
                {
                    dataIsNext = false;
                    continue;
                }
                else if (dataIsNext)
                {
                    string[] piecedResult = line.Split("^");
                    string ien = "";

                    if (piecedResult.Length > 1)
                    {
                        ien = piecedResult[0];

                        /* sometimes unique cross reference points to the same IEN */
                        if (listerResult.ContainsKey(ien))
                            continue; 
                        
                        Dictionary<string, string> listerRow = new Dictionary<string, string>();
                        listerResult.Add(ien, listerRow);

                        for (int i = 0; i < piecedResult.Length; i++)
                        {
                            listerRow.Add(listerMap[i], piecedResult[i]);
                        }
                    }
                }

            }
            
        }

        private void parseResponse()
        {
            string[] results = Response.Split("\r\n");
            bool mapIsNext = false;
            bool lookingForIens = true;
            bool dataIsNext = false;
            bool iensIsNext = false;

            HashSet<int> duplicateIenPositions = new HashSet<int>();
            Dictionary<int, string> ienMapping = new Dictionary<int, string>();
            int rowCount = 0;
            int rowIterator = 0;
            int fieldIterator = 0;

            for (int lineNumber = 0; lineNumber < results.Length; lineNumber++)
            {
                string line = results[lineNumber];

                if (line == "[MAP]")
                {
                    mapIsNext = true;
                    continue;
                }
                else if (mapIsNext)
                {
                    mapIsNext = false;

                    string[] piecedMap = line.Split("^");
                    for (int i = 0; i < piecedMap.Length; i++)
                    {
                        listerMap.Add(i, piecedMap[i]);
                    }

                    continue;
                }

                if (line == "[BEGIN_diDATA]")
                {
                    continue;
                }
                else if (line == "[END_diDATA]")
                {
                    continue;
                }

                if (lookingForIens && line == "BEGIN_IENs")
                {
                    lookingForIens = false;
                    iensIsNext = true;
                    rowIterator = 0;
                    rowCount = 0;
                    continue;
                }
                else if (lookingForIens == false && line == "END_IENs")
                {
                    iensIsNext = false;
                    dataIsNext = true;
                    continue;
                }
                else if(iensIsNext)
                {
                    rowCount++;

                    if (listerResult.ContainsKey(line))
                    {
                        duplicateIenPositions.Add(rowCount);
                    }
                    else
                    {
                        listerResult.Add(line, new Dictionary<string, string>());
                    }

                    ienMapping.Add(rowCount, line);

                    continue;
                }

                if (iensIsNext == false && line == "BEGIN_IDVALUES")
                {
                    dataIsNext = true;
                    rowIterator = 1;
                    fieldIterator = 0;
                    continue;
                }
                else if (dataIsNext == true && line == "END_IDVALUES")
                {
                    dataIsNext = false;
                    continue;
                }
                else if (dataIsNext)
                {
                    /**
                     * if there are duplicate iens - we must skip to the next record 
                     **/
                    if (duplicateIenPositions.Contains(rowIterator))
                    {
                        lineNumber = lineNumber + listerMap.Count - 1;
                        rowIterator++;
                        fieldIterator = 0;
                        continue;
                    }

                    Dictionary<string, string> row = listerResult[ienMapping[rowIterator]];
                    string fieldNumber = listerMap[fieldIterator];

                    row.Add(fieldNumber, line);

                    fieldIterator++;

                    if (fieldIterator >= listerMap.Count)
                    {
                        rowIterator++;
                        fieldIterator = 0;
                    }
                }
            }
        }

        private void parseErrors()
        {
            StringBuilder builderErrors = new StringBuilder();

            string[] results = Response.Split("\r\n");

            foreach (string line in results)
            {
                if (line == "[BEGIN_diERRORS]")
                {
                    continue;
                }
                else if (line == "[END_diERRORS]")
                {
                    if (builderErrors[builderErrors.Length - 1] == '\n')
                        builderErrors.Remove(builderErrors.Length - 1, 1);
                    throw new DdrListerException(builderErrors.ToString());
                }
                else
                {
                    builderErrors.Append(line);
                    builderErrors.Append("\n");
                }                
            }

            throw new DdrListerException("Error: [END_diERRORS] not found. " + builderErrors.ToString());
        }

        public Dictionary<int, string> GetMapping()
        {
            return listerMap;
        }

        public Dictionary<int, string>.ValueCollection GetMappingFields()
        {
            return listerMap.Values;
        }

        public Dictionary<string, Dictionary<string, string>> GetAllRows()
        {
            return listerResult;
        }

        public Dictionary<string, string> GetRowByIen(string ien)
        {
            return listerResult[ien];
        }

        public string GetValueByIenAndField(string ien, string field)
        {
            return listerResult[ien][field];
        }
    }
}
