﻿using System;
using System.Runtime.CompilerServices;
using System.Security;
using System.Timers;
using Microsoft.Extensions.Logging;
using Vista.RPC.Exceptions;
using Vista.RPC.Models;
using Vista.RPC.Procedures;
using Vista.RPC.Procedures.XUS;
using Vista.RPC.Procedures.XWB;

namespace Vista.RPC
{
    public class RpcClient : IRpcClient
    {
        private IRpcSocketFactory _rpcSocketFactory;
        private IRpcSocket _rpcSocket;
        private Timer _rpcTimer = new Timer();
        private ILogger _logger;

        public bool Connected => _rpcSocket != null && _rpcSocket.Connected;

        public string Hostname
        {
            get;
            protected set;
        }

        public int Port
        {
            get;
            protected set;
        }

        public bool PreventRpcTimeout
        {
            get; 
            protected set;
        }

        public double RpcReadTimeout
        {
            get; 
            protected set;
        }

		public IRpcSocketFactory RpcSocketFactory
		{
			set {
				_rpcSocketFactory = value;
				if (_rpcSocket.Connected)
				{
					_rpcSocket.Disconnect();
				}
				_rpcSocket = _rpcSocketFactory.Create();
			}
		}

        public RpcClient(ILogger logger, string vistaSiteId, bool preventRpcTimeout = true)
        {
            _logger = logger ?? throw new ArgumentNullException(nameof(logger));
            _logger.LogDebug($"rpcClient created, vistaSiteId: ${vistaSiteId}");

            _rpcSocketFactory = new RpcSocketFactory(logger as ILogger<IRpcSocket>);

			_rpcSocket = _rpcSocketFactory.Create();
            VistaSite vistaSite = RpcClientUtility.GetVistaSiteById(vistaSiteId);

            if (null != vistaSite)
            {
                Hostname = vistaSite.VistaHostname;
                Port = vistaSite.VistaPort;

                PreviousLogin = false;
                
                InitializeTimer(preventRpcTimeout);
            }
            else
            {
                var errorMessage = $"the VistA site could not be found: {vistaSiteId}";
                logger.LogError(errorMessage);
                throw new SiteNotFoundException(errorMessage);
            }
        }

        public RpcClient(ILogger logger, string hostname, int port, bool preventRpcTimeout = true) 
        {
            _logger = logger ?? throw new ArgumentNullException(nameof(logger));
            _logger.LogDebug($"rpcClient created, hostname: {hostname}, port: {port}");

            _rpcSocketFactory = new RpcSocketFactory(logger);

			_rpcSocket = _rpcSocketFactory.Create();
            if (!String.IsNullOrEmpty(hostname) && hostname.Length > 0)
            {
                Hostname = hostname;
            }

            if (port > 0 && port <= 65536)
            {
                Port = port;
            }

            PreviousLogin = false;
            
            InitializeTimer(preventRpcTimeout);
        }

        public RpcClient(ILogger logger, string hostname, string port, bool preventRpcTimeout = true)
        {
            _logger = logger ?? throw new ArgumentNullException(nameof(logger));
            _logger.LogDebug($"rpcClient created, hostname: {hostname}, port: {port}");

            _rpcSocketFactory = new RpcSocketFactory(logger as ILogger<IRpcSocket>);

			_rpcSocket = _rpcSocketFactory.Create();
            if (!String.IsNullOrEmpty(hostname) && hostname.Length > 0)
            {
                Hostname = hostname;
            }

            int portInt = -1;
            if (int.TryParse(port, out portInt))
            {
                if (portInt > 0 && portInt <= 65536)
                {
                    Port = portInt;
                }
            }            

            PreviousLogin = false;

            InitializeTimer(preventRpcTimeout);
        }

        private void InitializeTimer(bool preventRpcTimeout)
        {
            PreventRpcTimeout = preventRpcTimeout;
            _rpcTimer.Elapsed += rpcTimer_Elapsed;
        }

        public void Connect()
        {
            if (!Connected)
            {
                Connect remoteProcedure = new Connect(_rpcSocket.LocalIpAddress, _rpcSocket.LocalHostName);
                try
                {
                    _rpcSocket.Connect(Hostname, Port);
                    _rpcSocket.Call(remoteProcedure);
                }
                catch (RpcSocketException e)
                {
                    _logger.LogError(e, "failed to connect to the vista system");
                    throw new ConnectException(e.Message, e);
                }

                if (false == remoteProcedure.ConnectionAccepted())
                {
                    var errorMessage = "rpc broker is not accepting new style connections";
                    throw new ConnectException(errorMessage);
                }
            }

            BrokerInfo brokerInfo = new BrokerInfo();
            _rpcSocket.Call(brokerInfo);

            TimeSpan timeout = brokerInfo.GetTimeout();
            RpcReadTimeout = timeout.TotalSeconds;

            if (PreventRpcTimeout)
            {
                if (RpcReadTimeout > 20)
                {
                    _rpcTimer.Interval = 1000*(RpcReadTimeout - 20);
                }
                else
                {
                    _rpcTimer.Interval = 1000*RpcReadTimeout;
                }
            }

            TimerStart();
        }

        private void rpcTimer_Elapsed(object sender, ElapsedEventArgs e)
        {
            var sayHello = new RemoteProcedure("XWB IM HERE");
            Call(sayHello);
        }

        public bool PreviousLogin { get; private set; }
        private SecureString accessCode = null;
        private SecureString verifyCode = null;

        public SecureString CurrentContext { get; private set; }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="accessCode">unencrypted access code</param>
        /// <param name="verifyCode">unencrypted verify code</param>
        /// <exception cref="ConnectException">when the client is unable to connect</exception>
        /// <exception cref="RpcSocketException">when a socket is unable to read or write data</exception>
        public LoginResponse Login(SecureString accessCode, SecureString verifyCode)
        {
            /* step 1: connect to the socket */
            Connect(); 

            /* step 2: setup our m symbol table on server */
            Call(new SignonSetup());

            /* step 3: login using supplied a/v codes */
            AvCode avCode = new AvCode(accessCode, verifyCode);
            Call(avCode);

            var loginResponse = avCode.LoginResponse;

            if (false == loginResponse.LoggedIn)
            {
                this.PreviousLogin = false;
                this.accessCode = null;
                this.verifyCode = null;
            }
            else
            {
                this.accessCode = accessCode;
                this.verifyCode = verifyCode;
                this.PreviousLogin = true;
            }

            return loginResponse;
        }

        public LoginResponse Login(SecureString accessCode, SecureString verifyCode, SecureString brokerContext)
        {
            var loginResponse = Login(accessCode, verifyCode);

            if (!loginResponse.LoggedIn) return loginResponse;
            if (!CreateContext(brokerContext))
            {
                throw new CallException($"Context: {brokerContext.ToClearText()} is invalid");
            }

            return loginResponse;
        }

        public void ReLogin()
        {
            if (accessCode == null)
                throw new LoginException("you must first login successfully before attempting to re-login");

            if(CurrentContext == null)
            {
                Login(accessCode, verifyCode);
            }
            else
            {
                Login(accessCode, verifyCode, CurrentContext);
            }
        }

        public void Disconnect()
        {
            if (Connected)
            {
                try
                {
                    Disconnect disconnect = new Disconnect();
                    Call(disconnect);
                }
                finally
                {
                    _rpcSocket.Disconnect();
					/**
                     * due to the fact that closing the socket disposes the object when the
                     * socket is closed, we must create a new socket.
                     */
					_rpcSocket = _rpcSocketFactory.Create();
                }
            }

            TimerStop();
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="remoteProcedure"></param>
        /// <returns></returns>
        [MethodImpl(MethodImplOptions.Synchronized)]
        public void Call(IRemoteProcedure remoteProcedure)
        {
            TimerStop();

            if (!Connected)
            {
                throw new ConnectException("You must be connected to an RPC broker before calling a remote procedure");
            }

            /*if (PreviousLogin && !Connected)
            {
                ReLogin();
            }*/

            try
            {
                _rpcSocket.Call(remoteProcedure);
                remoteProcedure.AfterCall(); /* if available in the remote procedure - call the afterCall method */
            }
            catch (Exception e)
            {
                throw new RpcSocketException(e.Message, e);
            }
            finally
            {
                TimerStart();
            }
        }

        /// <summary>
        /// This method will call the remote procedure and then immediately disconnect
        /// </summary>
        /// <param name="rp"></param>
        public void CallAndDisconnect(IRemoteProcedure rp)
        {
            Call(rp);
            Done();
        }

        public void Done()
        {
            Disconnect();
        }

        public bool CreateContext(SecureString brokerContext)
        {
            CreateContext remoteProcedure = new CreateContext(brokerContext);
            Call(remoteProcedure);

            if (remoteProcedure.IsValidContext())
            {
                CurrentContext = brokerContext;
                return true;
            }

            return false;
        }

        public void Dispose()
        {
            Disconnect();
        }

        private void TimerStart()
        {
            if (PreventRpcTimeout)
            {
                _rpcTimer.Start();
            }
        }

        private void TimerStop()
        {
            if (PreventRpcTimeout)
            {
                _rpcTimer.Stop();
            }
        }
    }
}
