﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Xml.Linq;
using Vista.RPC.Models;

namespace Vista.RPC
{
    public class RpcClientUtility
    {

        public static readonly String VHA_SITES_FILE = "Resources\\VhaSites.xml";
        private static IEnumerable<XElement> vistaSites;

        /// <summary>
        /// Helper method - finds the VistA site information using the given VistA Site ID, for example
        /// the site id for Honolulu is 459 and the site id for Tampa is 673.
        /// </summary>
        /// <param name="vistaSiteId">The given site id to look up</param>
        /// <returns>VistaSite object otherwise null if the site id is not found</returns>
        public static VistaSite GetVistaSiteById(String vistaSiteId)
        {
            if (String.IsNullOrWhiteSpace(vistaSiteId))
            {
                return null;
            }

            XElement xElement = XElement.Load(VHA_SITES_FILE);

            vistaSites = from vistaSite in xElement.Elements().Elements()
                where (string) vistaSite.Attribute("ID").Value == vistaSiteId
                select vistaSite;

            return buildVistaSite(vistaSites);
        }

        /// <summary>
        /// Helper method - finds the VistA site information using the given VistA moniker, for example
        /// the moniker for Honolulu is HON and the moniker for Tampa is TAM.
        /// </summary>
        /// <param name="vistaMoniker">The given moniker to look up</param>
        /// <returns>VistaSite object otherwise null if the moniker is not found</returns>
        public static VistaSite GetVistaSiteByMoniker(String vistaMoniker)
        {
            if (String.IsNullOrWhiteSpace(vistaMoniker))
            {
                return null;
            }

            XElement xElement = XElement.Load(VHA_SITES_FILE);

            IEnumerable<XElement> xElements = from vistaSite in xElement.Elements().Elements()
                                              where (string)vistaSite.Attribute("moniker").Value == vistaMoniker
                                              select vistaSite;

            return buildVistaSite(xElements);
        }

        /// <summary>
        /// Builder method - takes the results of a LINQ query and returns a VistaSite object that contains the information
        /// </summary>
        /// <param name="linqXElements">resulting LINQ query</param>
        /// <returns>VistaSite if the LINQ query finds the results otherwise null</returns>
        private static VistaSite buildVistaSite(IEnumerable<XElement> linqXElements)
        {
            VistaSite vistaSite = new VistaSite();
            bool dataSourceFound = false;

            if(linqXElements.Count() <= 0 || linqXElements.Count() > 1)
                return null;

            XElement vistaSiteEntry = linqXElements.First();
            vistaSite.SiteName = vistaSiteEntry.Attribute("name").Value;
            vistaSite.SiteId = vistaSiteEntry.Attribute("ID").Value;
            vistaSite.SiteMoniker = vistaSiteEntry.Attribute("moniker").Value;

            foreach (var dataSource in linqXElements.Elements())
            {
                if (dataSource.Name.LocalName == "DataSource")
                {
                    if (dataSource.Attribute("modality").Value.ToUpper() == "HIS"
                        && dataSource.Attribute("protocol").Value.ToUpper() == "VISTA"
                        && dataSource.Attribute("status").Value.ToUpper() == "ACTIVE")
                    {
                        vistaSite.VistaHostname = dataSource.Attribute("source").Value.ToUpper();

                        // sometimes the MDWS VhaSites.xml will be missing the port information and throw an exception...
                        Int32.TryParse(dataSource.Attribute("port")?.Value, out vistaSite.VistaPort);

                        if (vistaSite.VistaPort == 0)
                        {
                            vistaSite.VistaPort = 9200;
                        }

                        dataSourceFound = true;
                    }
                }
            }                       

            if(dataSourceFound)
                return vistaSite;
            
            return null;
        }
    }
}
