﻿using System;
using System.Runtime.CompilerServices;
using System.Text;
using System.Net;
using System.Net.Sockets;
using Microsoft.Extensions.Logging;
using Vista.RPC.Procedures;

namespace Vista.RPC
{
    class RpcSocket: IRpcSocket
    {
        private readonly TcpClient tcpClient = new TcpClient();
        private readonly ILogger Logger;

        public RpcSocket(ILogger logger)
        {
            Logger = logger;
        }

        public void Connect(string hostname, int port)
        {
            lock (tcpClient)
            {
                tcpClient.Connect(hostname, port);
            }
        }

        public void Disconnect()
        {
            lock (tcpClient)
            {
                tcpClient.Close();
            }
        }

        public bool Connected
        {
            get
            {
                lock (tcpClient)
                {
                    if (null == tcpClient.Client)
                    {
                        return false;
                    }

                    return tcpClient.Connected;
                }
            }
        }

        /// <summary>
        /// i determine the local ip address once the socket is connected; better than examining the
        /// os tcp/ip routing table to determine which interface the packet will exit
        /// </summary>
        public string LocalIpAddress
        {
            get
            {
                if (!Connected)
                    return "";
                IPEndPoint endPoint = (IPEndPoint) tcpClient.Client.LocalEndPoint;
                return endPoint.Address.ToString();
            }
        }

        public string LocalHostName
        {
            get { return Dns.GetHostName(); }
        }

        [MethodImpl(MethodImplOptions.Synchronized)]
        public void Call(IRemoteProcedure rp)
        {
            lock (tcpClient)
            {
                ASCIIEncoding ascii = new ASCIIEncoding();
                byte[] buffer = ascii.GetBytes(rp.ToProcedure());

                /*
                 * Per the implementation, the NetworkStream will be closed when the TcpClient
                 * is disposed or closed (disonnected); therefore any uses of RpcClient shall
                 * execute within a using(...) {} statement.
                 *
                 * Source: https://referencesource.microsoft.com/#System/net/System/Net/Sockets/TCPClient.cs,455
                 */
                NetworkStream ns = tcpClient.GetStream();
                ns.Write(buffer, 0, buffer.Length);

                System.IO.MemoryStream ms = new System.IO.MemoryStream();

                int read = -2; /* -2 because its simply not -1 */
                int readMarker = 0;

                while ((read = ns.ReadByte()) != -1)
                {
                    if (4 == read)
                        break;

                    if (readMarker == 0 && read > 0)
                    {
                        readMarker++;
                        continue;
                    }
                    
                    if ((readMarker == 0 || readMarker == 1) && read == 0)
                    {
                        readMarker++;
                        continue;
                    }
                        
                    //if (nullHeader && 0 == read) /* weird beginning null characters? */
                    //    continue;

                    ms.WriteByte((byte)read);
                    readMarker++;
                }

                rp.Response = Encoding.UTF8.GetString(ms.ToArray());
            }
        }

        public void Dispose()
        {
            tcpClient?.Close();
            tcpClient?.Dispose();
        }
    }
}
