﻿using System;
using System.Runtime.InteropServices;
using System.Security;
using System.Security.Cryptography;
using System.Text;

namespace Vista.RPC
{
    public static class StringExtension
    {
        /// <summary>
        /// Helper extension; splits the string using the given delimiter
        /// </summary>
        /// <param name="value">the string to split</param>
        /// <param name="delimiter">how to split the string</param>
        /// <returns></returns>
        public static string[] Split(this String value, string delimiter)
        {
            return value.Split(new string[] { delimiter }, StringSplitOptions.None);
        }

        /// <summary>
        /// Helper extension; performs just like the MUMPS equivalent of $PIECE
        /// </summary>
        /// <param name="value">the value to $PIECE</param>
        /// <param name="delimiter">how to split the string</param>
        /// <param name="pieceNumber">first piece is 1 not a zero</param>
        /// <returns>the string of $P(value, delimiter, pieceNumber)</returns>
        public static string Piece(this String value, string delimiter, int pieceNumber)
        {
            string[] pieces = value.Split(delimiter);
            if (pieceNumber > pieces.Length)
            {
                return null;
            }

            return pieces[--pieceNumber];
        }

        /// <summary>
        /// Helper extension used by the RPC client
        /// </summary>
        /// <param name="value"></param>
        /// <param name="number"></param>
        /// <returns></returns>
        public static string PackString(this String value, int number)
        {
            StringBuilder sb = new StringBuilder(value.Length.ToString());

            while (sb.Length < number)
            {
                sb.Insert(0, "0");
            }

            sb.Append(value);

            return sb.ToString();
        }

        /// <summary>
        /// Helper extension; used by the RPC client
        /// </summary>
        /// <param name="value"></param>
        /// <returns></returns>
        public static string PackString(this String value)
        {
            // note value must be < 255
            StringBuilder sb = new StringBuilder();
            sb.Append(Convert.ToChar(value.Length));
            sb.Append(value);

            return sb.ToString();
        }

        /// <summary>
        /// Helper extension; used by the RPC client
        /// </summary>
        /// <param name="value"></param>
        /// <returns></returns>
        public static string PackVariable(this String value)
        {
            if (String.IsNullOrEmpty(value))
            {
                value = "0";
            }

            StringBuilder sb = new StringBuilder();
            sb.Append('|');
            sb.Append(Convert.ToChar(value.Length));
            sb.Append(value);

            return sb.ToString();
        }

        /// <summary>
        /// Helper extension; used by the RPC client
        /// </summary>
        /// <param name="value"></param>
        /// <param name="number"></param>
        /// <returns></returns>
        public static string PackLiteral(this String value, int number)
        {
            StringBuilder sb = new StringBuilder();
            sb.Append(value.Length.ToString());

            while (sb.Length < number)
            {
                sb.Insert(0, "0");
            }

            sb.Append(value);

            return sb.ToString();
        }

        /// <summary>
        /// VistA's cipher pad; this can be found in Z^XUSRB1.m on a VistA system
        /// </summary>
        private static readonly String[] CIPHER_PAD =
        {
            "wkEo-ZJt!dG)49K{nX1BS$vH<&:Myf*>Ae0jQW=;|#PsO`\'%+rmb[gpqN,l6/hFC@DcUa ]z~R}\"V\\iIxu?872.(TYL5_3",
            "rKv`R;M/9BqAF%&tSs#Vh)dO1DZP> *fX\'u[.4lY=-mg_ci802N7LTG<]!CWo:3?{+,5Q}(@jaExn$~p\\IyHwzU\"|k6Jeb",
            "\\pV(ZJk\"WQmCn!Y,y@1d+~8s?[lNMxgHEt=uw|X:qSLjAI*}6zoF{T3#;ca)/h5%`P4$r]G\'9e2if_>UDKb7<v0&- RBO.",
            "depjt3g4W)qD0V~NJar\\B \"?OYhcu[<Ms%Z`RIL_6:]AX-zG.#}$@vk7/5x&*m;(yb2Fn+l\'PwUof1K{9,|EQi>H=CT8S!",
            "NZW:1}K$byP;jk)7\'`x90B|cq@iSsEnu,(l-hf.&Y_?J#R]+voQXU8mrV[!p4tg~OMez CAaGFD6H53%L/dT2<*>\"{\\wI=",
            "vCiJ<oZ9|phXVNn)m K`t/SI%]A5qOWe\\&?;jT~M!fz1l>[D_0xR32c*4.P\"G{r7}E8wUgyudF+6-:B=$(sY,LkbHa#\'@Q",
            "hvMX,\'4Ty;[a8/{6l~F_V\"}qLI\\!@x(D7bRmUH]W15J%N0BYPkrs&9:$)Zj>u|zwQ=ieC-oGA.#?tfdcO3gp`S+En K2*<",
            "jd!W5[];4\'<C$/&x|rZ(k{>?ghBzIFN}fAK\"#`p_TqtD*1E37XGVs@0nmSe+Y6Qyo-aUu%i8c=H2vJ\\) R:MLb.9,wlO~P",
            "2ThtjEM+!=xXb)7,ZV{*ci3\"8@_l-HS69L>]\\AUF/Q%:qD?1~m(yvO0e\'<#o$p4dnIzKP|`NrkaGg.ufCRB[; sJYwW}5&",
            "vB\\5/zl-9y:Pj|=(R\'7QJI *&CTX\"p0]_3.idcuOefVU#omwNZ`$Fs?L+1Sk<,b)hM4A6[Y%aDrg@~KqEW8t>H};n!2xG{",
            "sFz0Bo@_HfnK>LR}qWXV+D6`Y28=4Cm~G/7-5A\\b9!a#rP.l&M$hc3ijQk;),TvUd<[:I\"u1\'NZSOw]*gxtE{eJp|y (?%",
            "M@,D}|LJyGO8`$*ZqH .j>c~h<d=fimszv[#-53F!+a;NC\'6T91IV?(0x&/{B)w\"]Q\\YUWprk4:ol%g2nE7teRKbAPuS_X",
            ".mjY#_0*H<B=Q+FML6]s;r2:e8R}[ic&KA 1w{)vV5d,$u\"~xD/Pg?IyfthO@CzWp%!`N4Z\'3-(o|J9XUE7k\\TlqSb>anG",
            "xVa1\']_GU<X`|\\NgM?LS9{\"jT%s$}y[nvtlefB2RKJW~(/cIDCPow4,>#zm+:5b@06O3Ap8=*7ZFY!H-uEQk; .q)i&rhd",
            "I]Jz7AG@QX.\"%3Lq>METUo{Pp_ |a6<0dYVSv8:b)~W9NK`(r\'4fs&wim\\kReC2hg=HOj$1B*/nxt,;c#y+![?lFuZ-5D}",
            "Rr(Ge6F Hx>q$m&C%M~Tn,:\"o\'tX/*yP.{lZ!YkiVhuw_<KE5a[;}W0gjsz3]@7cI2\\QN?f#4p|vb1OUBD9)=-LJA+d`S8",
            "I~k>y|m};d)-7DZ\"Fe/Y<B:xwojR,Vh]O0Sc[`$sg8GXE!1&Qrzp._W%TNK(=J 3i*2abuHA4C\'?Mv\\Pq{n#56LftUl@9+",
            "~A*>9 WidFN,1KsmwQ)GJM{I4:C%}#Ep(?HB/r;t.&U8o|l[\'Lg\"2hRDyZ5`nbf]qjc0!zS-TkYO<_=76a\\X@$Pe3+xVvu",
            "yYgjf\"5VdHc#uA,W1i+v\'6|@pr{n;DJ!8(btPGaQM.LT3oe?NB/&9>Z`-}02*%x<7lsqz4OS ~E$\\R]KI[:UwC_=h)kXmF",
            "5:iar.{YU7mBZR@-K|2 \"+~`M%8sq4JhPo<_X\\Sg3WC;Tuxz,fvEQ1p9=w}FAI&j/keD0c?)LN6OHV]lGy\'$*>nd[(tb!#"
        };

        /// <summary>
        /// Gets a random integer using the cryptographic random number generator with a min and max value
        /// </summary>
        /// <param name="minValue">the lower bound, inclusive</param>
        /// <param name="maxValue">the upper bound, inclusive</param>
        /// <returns></returns>
        public static int GetSecureRandomInt(Int32 minValue, Int32 maxValue)
        {
            var provider = new RNGCryptoServiceProvider();

            if (minValue > maxValue)
            {
                throw new ArgumentOutOfRangeException(nameof(minValue));
            }

            if (minValue == maxValue)
            {
                return minValue;
            }

            Int64 diff = maxValue - minValue;
            var byteArray = new byte[4];

            while (true)
            {
                provider.GetBytes(byteArray);
                UInt32 rand = BitConverter.ToUInt32(byteArray, 0);

                Int64 max = (1 + (Int64)UInt32.MaxValue);
                Int64 remainder = max % diff;
                if (rand < max - remainder)
                {
                    return (Int32)(minValue + (rand % diff));
                }
            }
        }

        /// <summary>
        /// Get a SecureString from a string
        /// </summary>
        /// <param name="value">The given string</param>
        /// <returns>a SecureString of the given string</returns>
        public static SecureString ToSecureString(this string value)
        {
            var secureString = new SecureString();
            foreach (char c in value)
            {
                secureString.AppendChar(c);
            }

            return secureString;
        }

        /// <summary>
        /// Get the unencrypted value of the SecureString in a String object
        /// </summary>
        /// <param name="value"></param>
        /// <returns>the </returns>
        public static String ToClearText(this SecureString value)
        {
            if (value == null)
            {
                return null;
            }

            if (value.Length == 0)
            {
                return String.Empty;
            }

            IntPtr pointer = IntPtr.Zero;

            var builder = new StringBuilder(value.Length);

            try
            {
                pointer = Marshal.SecureStringToGlobalAllocAnsi(value);
                for (int i = 0; i < value.Length; i++)
                {
                    builder.Append((char)Marshal.ReadByte(pointer, i));
                }
            }
            finally
            {
                if (pointer != IntPtr.Zero)
                {
                    Marshal.ZeroFreeGlobalAllocAnsi(pointer);
                    pointer = IntPtr.Zero;
                }
            }

            return builder.ToString();
        }

        /// <summary>
        /// Concat this SecureString with a bunch of other SecureStrings
        /// </summary>
        /// <param name="value">this SecureString</param>
        /// <param name="secureStrings">variable arguments of SecureString</param>
        /// <returns></returns>
        public static SecureString Concat(this SecureString value, params SecureString[] secureStrings)
        {
            IntPtr pointer = IntPtr.Zero;

            var secureStringResult = new SecureString();

            for (int i = 0; i <= secureStrings.Length; i++)
            {
                SecureString toExamine = null;

                toExamine = i == 0 ? value : secureStrings[i - 1];

                if (toExamine == null || toExamine.Length == 0)
                {
                    continue;
                }

                try
                {
                    pointer = Marshal.SecureStringToGlobalAllocAnsi(toExamine);
                    for (int j = 0; j < toExamine.Length; j++)
                    {
                        secureStringResult.AppendChar((char)Marshal.ReadByte(pointer, j));
                    }
                }
                finally
                {
                    if (pointer != IntPtr.Zero)
                    {
                        Marshal.ZeroFreeGlobalAllocAnsi(pointer);
                        pointer = IntPtr.Zero;
                    }
                }
            }

            return secureStringResult;
        }

        /// <summary>
        /// Encrypts the SecureString value using the VistA cipher
        /// </summary>
        /// <param name="value">the SecureString to encrypt</param>
        /// <returns>the Vista encrypted SecureString</returns>
        public static SecureString ToVistaEncrypt(this SecureString value)
        {
            if (value == null)
            {
                return null;
            }

            var toVistaEncrypt = new SecureString();

            IntPtr pointer = IntPtr.Zero;

            try
            {
                pointer = Marshal.SecureStringToGlobalAllocAnsi(value);
                
                const int MAX_KEY = 19; // CIPHER_PAD.Length - 1;

                int associatorIndex = 3; //GetSecureRandomInt(0, MAX_KEY);
                int identifierIndex = 0;

                do
                {
                    identifierIndex = GetSecureRandomInt(0, MAX_KEY);
                } while (associatorIndex == identifierIndex);

                identifierIndex = 4;

                toVistaEncrypt.AppendChar((char)(associatorIndex + 32));

                for (int i = 0; i < value.Length; i++)
                {
                    char inChar = (char) Marshal.ReadByte(pointer, i);
                    int pos = CIPHER_PAD[associatorIndex].IndexOf(inChar);
                    if (pos == -1)
                    {
                        toVistaEncrypt.AppendChar(inChar);
                    }
                    else
                    {
                        toVistaEncrypt.AppendChar(CIPHER_PAD[identifierIndex][pos]);
                    }
                }

                toVistaEncrypt.AppendChar((char)(identifierIndex + 32));
            }
            finally
            {
                if (pointer != IntPtr.Zero)
                {
                    Marshal.ZeroFreeGlobalAllocAnsi(pointer);
                    pointer = IntPtr.Zero;
                }
            }

            return toVistaEncrypt;
        }
    }
}
