﻿-- =============================================
-- Description:	Returns user's list of station roles (what role the user has at each station he/she can access)
-- =============================================
CREATE PROCEDURE [dbo].[uspGetUserStationRoles]
	-- Add the parameters for the stored procedure here

	@userId int,
	@applicationIds varchar(max)
AS
DECLARE @ACCESS_LEVEL_NATIONAL tinyint,
	    @ACCESS_LEVEL_REGION tinyint,
	    @ACCESS_LEVEL_VISN tinyint,
	    @ACCESS_LEVEL_STATION tinyint,
    	@ACCESS_STATUS_APPROVED varchar(10),
    	@ROLE_READONLY tinyint,
		@isNationalUser bit,
		@minAppID tinyint;

DECLARE @userRoleTable TABLE
		(
			RowID int,
			ApplicationID smallint,
			RegionID tinyint,
			VisnID tinyint,
			StationID smallint,
			StationName varchar(250),
			RoleID tinyint, 
			AccessLevelID tinyint,
			AccessStatus varchar(50),
			ShowPII bit
		);

DECLARE @applicationsTable table
		(
		ApplicationID tinyint
		);
DECLARE @rolesTable TABLE
		(
			RoleID tinyint,
			ShowPII bit
		);
BEGIN
	-- SET NOCOUNT ON added to prevent extra result sets from
	-- interfering with SELECT statements.
	SET NOCOUNT ON;

	BEGIN TRY
		--Set necessary variables
		SET @ACCESS_STATUS_APPROVED = 'APPROVED';
		SELECT @ROLE_READONLY = RoleID FROM Roles WHERE RoleName = 'Read Only';
		SELECT @ACCESS_LEVEL_NATIONAL = dbo.ufnGetAccessLevelConstant('National');
		SELECT @ACCESS_LEVEL_REGION = dbo.ufnGetAccessLevelConstant('Region');
		SELECT @ACCESS_LEVEL_VISN = dbo.ufnGetAccessLevelConstant('Visn');
		SELECT @ACCESS_LEVEL_STATION = dbo.ufnGetAccessLevelConstant('Station');
			--will need to start the wrapper here for all permutations of appID--
			--build a table of various appId's--
			IF @applicationIds IS NULL
				BEGIN				
					INSERT INTO @applicationsTable (ApplicationID)
					SELECT ApplicationID FROM Applications
				END
			ELSE
				BEGIN
					INSERT INTO @applicationsTable (ApplicationID)
					SELECT [Value] FROM ufnSplit(',', @applicationIds)
				END
			--Set the min rowId to start the while loop
			SELECT @minAppID = MIN(ApplicationID)
			FROM @applicationsTable;

	   WHILE @minAppID IS NOT NULL 
		BEGIN
			--Clear out the @rolesTable
			DELETE FROM @rolesTable

			SELECT @isNationalUser = dbo.ufnIsNationalUser(@userId, @minAppID);
			--If is a national user or no user id is returned, then return all
			IF @userId Is Null OR @userId = 0 OR @isNationalUser = 1
				BEGIN

					IF EXISTS(SELECT RoleID FROM UserRoles WHERE UserID = @userId AND ApplicationID = @minAppID AND AccessLevelID = @ACCESS_LEVEL_NATIONAL 
					AND AccessStatus = @ACCESS_STATUS_APPROVED)
						BEGIN
							INSERT INTO @rolesTable (RoleID, ShowPII)
							SELECT DISTINCT RoleID, ShowPII 
								FROM UserRoles 
								WHERE UserID = @userId 
									AND ApplicationID = @minAppID
									AND AccessLevelID = @ACCESS_LEVEL_NATIONAL 
									AND AccessStatus = @ACCESS_STATUS_APPROVED;

						END
						--If a role isn't found, then substitute read only
						IF (SELECT COUNT(1) FROM @rolesTable) = 0
							BEGIN
								INSERT INTO @rolesTable (RoleID, ShowPII)
								VALUES (@ROLE_READONLY, 0)
							END	

						--Insert the possible station combinations into the table variable
						--not sure if applicationId needs to be called here via join-josh--
						INSERT INTO @userRoleTable (RowID, ApplicationID, RegionID, VisnID, StationID, StationName, 
							RoleID, AccessLevelID, AccessStatus, ShowPII)
						SELECT DISTINCT CAST(ROW_NUMBER() OVER(ORDER BY t.StationID) as int), 
							@minAppID, v.RegionID, t.VisnID, t.StationID, 
							'V' + VisnCode + ': ' + t.Station3N + ': ' + t.StationName, 
							RoleID, @ACCESS_LEVEL_NATIONAL, @ACCESS_STATUS_APPROVED, ShowPII
							FROM Stations t
							INNER JOIN Visns v ON t.VisnID = v.VisnID
							CROSS APPLY (
								SELECT RoleID, ShowPII
								FROM @rolesTable
								) as rx
						
					END
			--In cases where a user might be a national administrator with ShowPII = false, but have ShowPII = true for
			--a lesser user role, it is necessary to join the national admin and the other possible roles together
			--note this only happens for national level people.
			--ELSE
			BEGIN
				INSERT INTO @userRoleTable (ApplicationID, RegionID, VisnID, StationID, StationName, RoleID, AccessStatus, AccessLevelID, ShowPII)
					--check Region permissions
					SELECT DISTINCT ApplicationID, v.RegionID, t.VisnID, t.StationID, 
						CAST(t.VisnID as varchar) + ': ' + t.Station3N + ': ' + t.StationName, RoleID, AccessStatus, AccessLevelID, ShowPII
						FROM Stations t 
						INNER JOIN Visns v ON t.VisnID = v.VisnID
						INNER JOIN UserRoles ON v.RegionID = UserRoles.RegionID
						WHERE UserID = @userId 
						AND ApplicationID = @minAppID
						AND AccessLevelID = @ACCESS_LEVEL_REGION
						AND AccessStatus = @ACCESS_STATUS_APPROVED
					--check Visn permissions
					UNION 
						SELECT DISTINCT ApplicationID, v.RegionID, t.VisnID, t.StationID, 
							CAST(t.VisnID as varchar) + ': ' + t.Station3N + ': ' + t.StationName, RoleID, AccessStatus, AccessLevelID, ShowPII
							FROM Stations t 
							INNER JOIN Visns v ON t.VisnID = v.VisnID
							INNER JOIN UserRoles ON v.VisnID = UserRoles.VisnID
							WHERE UserID = @userId 
							AND ApplicationID = @minAppID
							AND AccessLevelID = @ACCESS_LEVEL_VISN
							AND AccessStatus = @ACCESS_STATUS_APPROVED
					
					--check Station permissions
					UNION 
						SELECT DISTINCT ApplicationID, v.RegionID, t.VisnID, t.StationID, 
							CAST(t.VisnID as varchar) + ': ' + t.Station3N + ': ' + t.StationName, RoleID, AccessStatus, AccessLevelID, ShowPII
							FROM Stations t 
							INNER JOIN Visns v ON t.VisnID = v.VisnID
							INNER JOIN UserRoles ON t.StationID = UserRoles.StationID
							WHERE UserID = @userId 
							AND ApplicationID = @minAppID
							AND AccessLevelID = @ACCESS_LEVEL_STATION
							AND AccessStatus = @ACCESS_STATUS_APPROVED


				END
				--Iterate at the end of the loop to the next RowID
				SELECT @minAppID = MIN(ApplicationID) 
				FROM @applicationsTable 
				WHERE ApplicationID > @minAppID ;

			END
				--Need to spoof unnecessary columns in order to hook into the data domain for UserRoles	
				SELECT CAST(ROW_NUMBER() OVER(ORDER BY StationID) AS int) AS 'RowID', 
					ApplicationID as 'ApplicationID',
					VisnID AS 'VisnID', 
					StationID AS 'StationID',
					RoleID AS 'RoleID', 
					AccessStatus AS 'AccessStatus', 
					AccessLevelID AS 'AccessLevelID', 
					ShowPII AS 'ShowPII',
					@userId as 'UserID'
					FROM @userRoleTable;

	END TRY

	BEGIN CATCH
		PRINT 'There was an error processing the user station roles' + CHAR(13);
		SELECT * FROM dbo.ufnGetErrorInfo();
		EXEC uspRaiseErrorInfo;
	END CATCH
END
GO

