﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Data.Entity;
using System.Linq.Expressions;
using WebVRAM.Data.Domain;

namespace WebVRAM.Data.Repository
{

    public class GenericRepository<TEntity> where TEntity : class
    {
        internal WebVRAM_Entities Context;
        internal DbSet<TEntity> DbSet;

        public GenericRepository()
        {
        }

        public GenericRepository(WebVRAM_Entities context)
        {
            this.Context = context;
            this.DbSet = context.Set<TEntity>();
        }


        /// <summary>
        /// Gets object or object collection, returns a list even if one object
        /// </summary>
        /// <param name="filter"></param>
        /// <param name="orderBy"></param>
        /// <param name="includeProperties"></param>
        /// <returns></returns>
        public virtual IEnumerable<TEntity> Get(
            Expression<Func<TEntity, bool>> filter = null,
            Func<IQueryable<TEntity>, IOrderedQueryable<TEntity>> orderBy = null,
            string includeProperties = "")
        {
            IQueryable<TEntity> query = DbSet;

            if (filter != null)
            {
                query = query.Where(filter);
            }

            foreach (var includeProperty in includeProperties.Split
                (new char[] { ',' }, StringSplitOptions.RemoveEmptyEntries))
            {
                query = query.Include(includeProperty);
            }

            if (orderBy != null)
            {
                return orderBy(query).ToList();
            }
            else
            {
                return query.ToList();
            }
        }

        /// <summary>
        /// Get record by id for the defined object type
        /// </summary>
        /// <param name="id"></param>
        /// <returns></returns>
        public virtual TEntity GetById(object id)
        {
            return DbSet.Find(id);
        }

        /// <summary>
        /// Executes ad hoc sql using SqlQuery method
        /// </summary>
        /// <param name="query"></param>
        /// <param name="parameters"></param>
        /// <returns></returns>
        public virtual IEnumerable<TEntity> GetObjectsBySQLQuery(string query, params object[] parameters)
        {
            return DbSet.SqlQuery(query, parameters).ToList();
        }


        /// <summary>
        /// Add new record for supplied object
        /// </summary>
        /// <param name="entity"></param>
        public virtual void Insert(TEntity entity)
        {
            this.DbSet.Add(entity);
        }

        /// <summary>
        /// Update existing record for supplied object
        /// </summary>
        /// <param name="entity"></param>
        public virtual void Update(TEntity entity)
        {
            if (this.Context.Entry(entity).State == EntityState.Detached)
            {
                this.DbSet.Attach(entity);
            }

            this.Context.Entry(entity).State = EntityState.Modified;
        }

        public virtual bool SendSQLUpdate(string sql)
        {
            IEnumerable<TEntity> result = DbSet.SqlQuery(sql);
            if (result == null) return false;
            return true;
        }

        /// <summary>
        /// Delete single record for supplied object 
        /// </summary>
        /// <param name="entity"></param>
        public virtual void Delete(TEntity entity)
        {
            if (this.Context.Entry(entity).State == EntityState.Detached)
            {
                this.DbSet.Attach(entity);
            }
            this.DbSet.Remove(entity);
        }

        /// <summary>
        /// Delete a specific record by id provided
        /// </summary>
        /// <param name="where"></param>
        public virtual void Delete(object id)
        {
            TEntity entityToDelete = this.DbSet.Find(id);
            Delete(entityToDelete);
        }

        /// <summary>
        /// Delete records based on WHERE clause criteria for defined object type
        /// </summary>
        /// <param name="where"></param>
        public virtual void Delete(Expression<Func<TEntity, bool>> where)
        {
            IEnumerable<TEntity> objects = this.DbSet.Where<TEntity>(where).AsEnumerable();
            foreach (TEntity obj in objects)
            {
                this.DbSet.Remove(obj);
            }

        }

        /// <summary>
        /// Allows for sorting any object passed in by the expression provided
        /// </summary>
        /// <typeparam name="TEntity"></typeparam>
        /// <typeparam name="TType"></typeparam>
        public class SortExpression<TEntity, TType>
        {
            Expression<Func<TEntity, TType>> SortProperty;
        }
    }
}
