﻿using System.Threading.Tasks;
using System;
using WebVRAM.Data.Domain;

namespace WebVRAM.Data.Repository
{
    public class UnitOfWork : IDisposable, IUnitOfWork
    {
        #region Class Properties

        private WebVRAM_Entities Context { get; set; }

        private UserRepository _userRepository;

        private UserRolesRepository _userRolesRepository;

        private WebLogsRepository _webLogsRepository;

        private VistaUserRepository _vistaUserRepository;

        private StationsRepository _stationsRepository;

        private bool Disposed { get; set; }

        #endregion

        #region Constructors
        
        /// <summary>
        /// Creates a UnitOfWork using the given entity factory
        /// </summary>
        /// <param name="entitiesFactory">the factory that will build the entity</param>
        public UnitOfWork(WebVRAM_EntitiesFactory entitiesFactory)
        {
            Context = entitiesFactory.Create();
            Disposed = false;
        }

        /// <summary>
        /// Creates a UnitOfWork using the given entities
        /// </summary>
        /// <param name="entities">the entities to be used by the UnitOfWork</param>
        public UnitOfWork(WebVRAM_Entities entities)
        {
            Context = entities;
            Disposed = false;
        }

        #endregion

        #region Repositories

        #region User Repository Property

        public UserRepository UserRepository
        {
            get { return this._userRepository ?? (this._userRepository = new UserRepository(Context)); }
            set { this._userRepository = value; }
        }

        #endregion

        #region User Roles Repository Property

        public UserRolesRepository UserRolesRepository
        {
            get { return this._userRolesRepository ?? (this._userRolesRepository = new UserRolesRepository(Context)); }
            set { this._userRolesRepository = value; }
        }

        #endregion        

        #region Web Logs Repository

        public WebLogsRepository WebLogsRepository
        {
            get { return _webLogsRepository ?? (_webLogsRepository = new WebLogsRepository(Context)); }
            set { _webLogsRepository = value; }
        }
        
        #endregion

        #region Vista User Repository

        public VistaUserRepository VistaUserRepository
        {
            get { return _vistaUserRepository ?? (_vistaUserRepository = new VistaUserRepository(Context)); }
            set { _vistaUserRepository = value; }
        }

        #endregion

        #region Stations Repository

        public StationsRepository StationsRepository
        {
            get { return _stationsRepository ?? (_stationsRepository = new StationsRepository(Context)); }
            set { _stationsRepository = value; }
        }

        #endregion

        #endregion

        #region Main Methods

        ///<summary>Method to get Generic Repository</summary>
        ///
        public GenericRepository<T> GetRepository<T>() where T : class
        {
            return new GenericRepository<T>(Context);
        }

        public int Save()
        {
            int val = 0;
            try
            {
                val = Context.SaveChanges();
            }
            catch (Exception ex)
            {
                if (ex.Message.ToLower().Contains("key") && ex.Message.ToLower().Contains("violation"))
                {
                    throw new Exception(
                        "This record has already been entered in the database. The same record cannot be entered new twice.");
                }
                else throw ex;
            }
            return val;
        }

        public Task<int> SaveAsync()
        {
            return Context.SaveChangesAsync();
        }


        protected virtual void Dispose(bool disposing)
        {
            if (!this.Disposed)
            {
                if (disposing)
                {
                    Context.Dispose();
                }
            }
            this.Disposed = true;
        }

        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        #endregion
    }
}