﻿using System;
using System.Collections.Generic;
using System.Linq;
using WebVRAM.Data.Domain;

namespace WebVRAM.Data.Repository
{
    public class UserRepository : GenericRepository<User>
    {
        public UserRepository(WebVRAM_Entities context)
        {
            Context = context;
            DbSet = context.Set<User>();
        }

        /// <summary>
        /// Get User By Portion of Last or First Name
        /// </summary>
        /// <param name="namePart"></param>
        /// <returns></returns>
        public List<User> GetUsersByNamePart(string namePart)
        {
            return
                (from user in Context.Users
                    where user.FirstName.Contains(namePart) || user.LastName.Contains(namePart)
                    select user).ToList<User>();
        }

        /// <summary>
        /// Check if user already exists in database
        /// </summary>
        /// <param name="emailAddress">string emailAddress</param>
        /// <returns>boolean</returns>
        public bool CheckIfUserExistsInDb(string emailAddress)
        {
            var existingUser =
                (from user in Context.Users
                    where String.Equals(user.EmailAddress, emailAddress, StringComparison.CurrentCultureIgnoreCase)
                    select user).SingleOrDefault();

            return (existingUser != null);
        }

        /// <summary>
        /// Get a User by providing a Windows AD user name
        /// </summary>
        /// <param name="userName">Windows AD username, given vha21\vhahonuyehac - the username is vhahonuyehac</param>
        /// <returns>a User entity or null if not found</returns>
        public User GetUserByUserName(string userName)
        {
            return Context.Users.FirstOrDefault(u => u.UserName == userName);
        }

        /// <summary>
        /// Get a list of active directory domains
        /// </summary>
        /// <returns></returns>
        public List<string> GetAdDomains()
        {
            return Context.uspGetADDomains().ToList();
        }        

        /// <summary>
        /// Get a list of users for a given active directory domain
        /// </summary>
        /// <param name="adDomain"></param>
        /// <param name="orderBy"></param>
        /// <param name="sortDirection"></param>
        /// <param name="page"></param>
        /// <param name="pageSize"></param>
        /// <returns></returns>
        public IQueryable<User> GetUsersByAdDomain(string adDomain, string orderBy, string sortDirection, int? page,
            int? pageSize)
        {
            var users =
                (from user in Context.Users where user.ADDomain.ToLower().Equals(adDomain.ToLower()) select user);

//            if (!string.IsNullOrEmpty(orderBy))
//                users = users.OrderBy(orderBy,
//                    (sortDirection.ToUpper() == "DESCENDING") ? SortDirection.Descending : SortDirection.Ascending);

            users = users.OrderBy(order => order.LastName);

            if (page.HasValue && pageSize.HasValue)
            {
                users = users.Skip((page.Value - 1)*pageSize.Value).Take(pageSize.Value);
            }

            return users;
        }

        public IQueryable<User> GetLastTenUsers()
        {
            return (from user in Context.Users orderby user.UserID descending select user).Take(10);
        }

        public IQueryable<User> SearchUsers(string firstName = "", string lastName = "", string email = "", 
            string domain = "", string username = "")
        {
            var query = Context.Users.AsQueryable();

            if (!string.IsNullOrEmpty(firstName))
                query = query.Where(u => u.FirstName.Contains(firstName));

            if (!string.IsNullOrEmpty(lastName))
                query = query.Where(u => u.LastName.Contains(lastName));

            if (!string.IsNullOrEmpty(email))
                query = query.Where(u => u.EmailAddress.Contains(email));

            if (!string.IsNullOrEmpty(domain))
                query = query.Where(u => u.ADDomain == domain);

            if (!string.IsNullOrEmpty(username))
                query = query.Where(u => u.UserName.Contains(username));

            return query;
        }

        public List<uspGetUserRegions_Result> GetUserRegions(int userId)
        {
            return Context.uspGetUserRegions(userId, "1").ToList();
        }

        public List<uspGetUserVisns_Result> GetUserVisns(int userId)
        {
            return Context.uspGetUserVisns(userId, "1").ToList();
        }

        public List<uspGetUserParentStations_Result> GetUserParentStations(int userId)
        {
            return Context.uspGetUserParentStations(userId, "1").ToList();
        }


        public bool IsUserActive(string userName)
        {
            var queryable = Context.Users.Where(u => u.UserName == userName && u.Inactive == false);
            return queryable.FirstOrDefault() != null;
        }
    }
}