using System;
using System.Linq;
using System.Security;
using System.Security.Claims;
using Microsoft.Extensions.Logging;
using Vista.RPC;
using Vista.RPC.Exceptions;
using WebVRAM.Web.ng.Common;
using WebVRAM.Web.ng.Models;

namespace WebVRAM.Vista.Services
{
    public class BaseService<T>
    {
        protected ILogger<T> Logger;
        protected IRpcClientFactory RpcClientFactory;
        protected ClaimsPrincipal User { get; private set; }

        private VistaConfig _vistaConfig;

        public BaseService(ILogger<T> logger, VistaConfig vistaConfig, ClaimsPrincipal user)
        {
            Logger = logger;
            RpcClientFactory = new RpcClientFactory(logger);
            User = user;

            _vistaConfig = vistaConfig;
        }

        protected string ClaimsHostname => _vistaConfig.ClaimsHostname;
        protected int ClaimsPort => _vistaConfig.ClaimsPort;
        protected SecureString VistaVramMenuOption => _vistaConfig.VistaMenuOption;
        protected SecureString BseApplicationPassPhrase => _vistaConfig.VistaBsePassPhrase;

        protected string UserIen
        {
            get
            {
                if (User.HasClaim(c => c.Type == JwtPrivateClaims.ClaimsDUZ))
                {
                    return User.Claims.FirstOrDefault(c => c.Type == JwtPrivateClaims.ClaimsDUZ)?.Value;
                }

                return null;
            }
        }

        protected SecureString BseToken
        {
            get
            {
                if (User.HasClaim(c => c.Type == JwtPrivateClaims.BseToken))
                {
                    return User.Claims.FirstOrDefault(c => c.Type == JwtPrivateClaims.BseToken)?.Value.ToSecureString();
                }

                return null;
            }
        }

        protected IRpcClient GetRpcClient(bool preventRpcTimeout = false, bool tryToLogin = true,
            string vistaHostname = null, int vistaPort = 0)
        {
            if (!tryToLogin)
            {
                if (string.IsNullOrEmpty(vistaHostname))
                {
                    return RpcClientFactory.Create(ClaimsHostname, ClaimsPort, preventRpcTimeout);
                }

                if (vistaPort <= 0)
                {
                    var errorMessage = "vistaPort should be greater than zero";
                    Logger.LogError(errorMessage);
                    throw new ArgumentException(errorMessage);
                }

                return RpcClientFactory.Create(vistaHostname, vistaPort, preventRpcTimeout);
            }

            if (!User.Identity.IsAuthenticated)
            {
                Logger.LogCritical("unauthenticated user attempting to connect");
                return null;
            }

            if (!User.HasClaim(c => c.Type == JwtPrivateClaims.AccessCode) ||
                !User.HasClaim(c => c.Type == JwtPrivateClaims.VerifyCode))
            {
                Logger.LogError("user is authenticated but doesn't have a valid a/v code pair");
                return null;
            }

            var accessCode = User.Claims.FirstOrDefault(c => c.Type == JwtPrivateClaims.AccessCode)?.Value.ToSecureString();
            var verifyCode = User.Claims.FirstOrDefault(c => c.Type == JwtPrivateClaims.VerifyCode)?.Value.ToSecureString();

            try
            {
                IRpcClient rpcClient;

                if (string.IsNullOrEmpty(vistaHostname))
                {
                    vistaHostname = ClaimsHostname;
                    vistaPort = ClaimsPort;
                }

                rpcClient = RpcClientFactory.Create(vistaHostname, vistaPort, preventRpcTimeout);

                var loginResponse = rpcClient.Login(accessCode, verifyCode, VistaVramMenuOption);

                if (loginResponse.LoggedIn == false)
                    return null;

                return rpcClient;
            }
            catch (RpcSocketException e)
            {
                Logger.LogError(e, "failed to access the vista system");
            }

            return null;
        }

        protected IRpcClient GetRpcClient()
        {
            return GetRpcClient(false);
        }
    }
}