using System;
using System.Security;
using System.Security.Claims;
using Microsoft.Extensions.Logging;
using Vista.RPC;
using Vista.RPC.Exceptions;
using Vista.RPC.Parameters;
using Vista.RPC.Procedures;
using Vista.RPC.Procedures.DDR;
using Vista.RPC.Procedures.XUS;
using WebVRAM.Vista.Service.Models;
using WebVRAM.Web.ng.Models;

namespace WebVRAM.Vista.Services
{
    public class LoginService : BaseService<LoginService>
    {
        public LoginService(ILogger<LoginService> logger, VistaConfig vistaConfig,
            ClaimsPrincipal claimsPrincipal) : base(logger, vistaConfig, claimsPrincipal)
        {
        }

        public string GetClaimsWelcomeMessage()
        {
            using (var rpcClient = GetRpcClient(tryToLogin: false))
            {
                try
                {
                    var rpc = new IntroMsg();

                    rpcClient.Connect();
                    rpcClient.CallAndDisconnect(rpc);

                    return rpc.Response;
                }
                catch (RpcSocketException e)
                {
                    Logger.LogError(e, "unable to fetch vista welcome message. vista offline?");
                    return null;
                }
            }
        }
        
        public LoginResult Login(SecureString accessCode, SecureString verifyCode = null)
        {
            // access code cannot be null
            if(accessCode == null)
                throw new ArgumentNullException(nameof(accessCode));

            // setup empty return value
            var loginResult = new LoginResult();

            // get and use an rpc client
            using (var rpcClient = GetRpcClient(tryToLogin: false))
            {
                try
                {
                    // perform login
                    var loginStatus = rpcClient.Login(accessCode, verifyCode, VistaVramMenuOption);

                    // bug - when verify code must be changed this fails
                    if (loginStatus.VerifyNeedsChanging)
                    {
                        return new LoginResult()
                        {
                            LoggedIn = false,
                            VerifyNeedsChanging = true
                        };
                    }

                    if (loginStatus.LoggedIn == false)
                    {
                        var errorMessage = loginStatus.ErrorMessage.ToLower();

//                    if (errorMessage.Contains("not a valid access code"))
//                        throw new InvalidAvCodeException(loginStatus.ErrorMessage);
//                    if (errorMessage.Contains("is locked"))
//                        throw new UserLockedException(loginStatus.ErrorMessage);
//                    if (errorMessage.Contains("no access allowed"))
//                        throw new UserDisabledException(loginStatus.ErrorMessage);

                        return new LoginResult()
                        {
                            LoggedIn = false,
                            ErrorMessage = loginStatus.ErrorMessage
                        };
                    }

                    loginResult = new LoginResult()
                    {
                        LoggedIn = loginStatus.LoggedIn,
                        LoggedInDuz = loginStatus.LoggedInDuz,
                        ErrorMessage = loginStatus.ErrorMessage,
                        LoginMessage = loginStatus.LoginMessage,
                        VerifyNeedsChanging = loginStatus.VerifyNeedsChanging
                    };

                    // get the bse token
                    InitializeBse(rpcClient, loginResult);

                    // get further details
                    GetLoginDetails(rpcClient, loginResult);
                }
                catch (ConnectException e)
                {
                    Logger.LogError(e, "rpc connection error");
                }
                catch (RpcSocketException e)
                {
                    Logger.LogError(e, "rpc socket error");
                }
                finally
                {
                    rpcClient.Disconnect();
                }
            }

            return loginResult;
        }
        
        private string InitializeBse(IRpcClient rpcClient, LoginResult loginResult)
        {
            if (!rpcClient.Connected)
            {
                var errorMessage = "rpc client should not be disconnected";
                Logger.LogError(errorMessage);
                throw new InvalidOperationException(errorMessage);
            }

            var divisionGet = new DivisionGet();
            rpcClient.Call(divisionGet);

            var visitorSet = new VisitorSet();
            rpcClient.Call(visitorSet);

            loginResult.BseToken = visitorSet.GetVisitorToken();

            return visitorSet.GetVisitorToken();
        }
        
        private void GetLoginDetails(IRpcClient rpcClient, LoginResult loginResult)
        {
            if (!rpcClient.Connected)
            {
                var errorMessage = "rpc client should not be disconnected";
                Logger.LogError(errorMessage);
                throw new InvalidOperationException(errorMessage);
            }
            
            const string newPersonFields = ".01;8;20.4;20.2;20.3";
            var userIen = loginResult.LoggedInDuz;
            var ddrGets = new DdrGets("200", userIen + ",", newPersonFields , "IE");
            rpcClient.Call(ddrGets);

            loginResult.PersonName = ddrGets.GetFileManFieldValue(".01").ExternalValue;
            loginResult.Title = ddrGets.GetFileManFieldValue("8").ExternalValue;

            loginResult.ElectronicSignature = ddrGets.GetFileManFieldValue("20.4").InternalValue;
            loginResult.ElectronicSignaturePrintedBlock = ddrGets.GetFileManFieldValue("20.2").ExternalValue;
            loginResult.ElectronicSignatureTitleBlock = ddrGets.GetFileManFieldValue("20.3").ExternalValue;
        }
        
        public ChangeVerifyCodeResult ChangeVerifyCode(SecureString accessCode, SecureString oldVerifyCode, SecureString newVerifyCode)
        {
            // access code cannot be null
            if (accessCode == null)
                throw new ArgumentNullException(nameof(accessCode));

            // get and use an rpc client
            using (var rpcClient = GetRpcClient())
            {
                try
                {
                    // perform login
                    var loginStatus = rpcClient.Login(accessCode, oldVerifyCode, VistaVramMenuOption);

                    var remoteProcedure = new RemoteProcedure("XUS CVC");
                    var verifyCodeParameter = (oldVerifyCode.ToVistaEncrypt() + "^" + newVerifyCode.ToVistaEncrypt() +
                                               "^" +
                                               newVerifyCode.ToVistaEncrypt()).ToSecureString();

                    remoteProcedure.AddParameter(new LiteralParameter(verifyCodeParameter));

                    rpcClient.Call(remoteProcedure);

                    if (remoteProcedure.Response.Trim().StartsWith("0"))
                        return new ChangeVerifyCodeResult() {VerifyCodeChanged = true};

                    var splitResponse = remoteProcedure.Response.Trim().Split("\r\n");
                    if (splitResponse.Length > 1)
                    {
                        return new ChangeVerifyCodeResult()
                        {
                            VerifyCodeChanged = false,
                            ErrorMessage = splitResponse[1]
                        };
                    }
                }
                catch (ConnectException e)
                {
                    Logger.LogError(e, "rpc connection error");
                }
                catch (RpcSocketException e)
                {
                    Logger.LogError(e, "rpc socket error");
                }
                finally
                {
                    rpcClient.Disconnect();
                }

                return new ChangeVerifyCodeResult()
                {
                    VerifyCodeChanged = false // TODO examine why this wrong
                };
            }
        }
    }
}