using System;
using System.Collections.Generic;
using System.Linq;
using System.Security.Claims;
using System.Threading.Tasks;
using Microsoft.Extensions.Logging;
using Vista.RPC;
using Vista.RPC.Exceptions;
using Vista.RPC.Procedures.DDR;
using Vista.RPC.Procedures.XU;
using Vista.RPC.Procedures.XUS;
using Vista.RPC.Utilities;
using WebVRAM.Vista.Service.Exceptions;
using WebVRAM.Vista.Service.Models;
using WebVRAM.Web.ng.Models;

namespace WebVRAM.Vista.Services
{
    /// <summary>
    /// This VistA service is the core of WebVRAM; it provides the synchronization logic and rules
    /// </summary>
    public class SyncService : BaseService<SyncService>
    {
        private VistaUserService _vistaUserService;

        public SyncService(ILogger<SyncService> logger, VistaConfig vistaConfig,
            ClaimsPrincipal claimsPrincipal, VistaUserService vistaUserService) : base(logger, vistaConfig, claimsPrincipal)
        {
            _vistaUserService = vistaUserService;
        }

        /// <summary>
        /// Gets a set of VistA security keys that are NOT allowed to be copied during synchronization
        /// </summary>
        /// <returns>a set of forbidden security keys</returns>
        private HashSet<string> GetForbiddenKeys()
        {
            using (var rpcClient = GetRpcClient())
            {
                // get all forbidden security keys
                var forbiddenKeysLister = new DdrLister(file: "100100.0344", fields: ".01");
                rpcClient.Call(forbiddenKeysLister);
                var allForbiddenKeys = forbiddenKeysLister.GetAllRows();

                var forbiddenKeys = new HashSet<string>();

                foreach (var row in allForbiddenKeys)
                {
                    var rowMap = row.Value;
                    forbiddenKeys.Add(rowMap[".01"]);
                }

                return forbiddenKeys;
            }
        }

        /// <summary>
        /// The progress dialog used to update the browser gui
        /// </summary>
        private IProgress<ProgressDialog> _progress;

        /// <summary>
        /// Helper method to update a progress dialog without repetitious code
        /// </summary>
        /// <param name="message">the progress dialog message to show</param>
        /// <param name="percentComplete">the progress bar percentage to show</param>
        /// <returns>a task</returns>
        private async Task UpdateProgress(string message, int percentComplete)
        {
            if (_progress != null)
            {
                _progress.Report(new ProgressDialog()
                {
                    Message = message,
                    PercentComplete = percentComplete
                });
                await Task.Delay(20);
            }
        }

        /// <summary>
        /// Synchronize the user to a single VistA site
        /// </summary>
        /// <param name="vistaSiteIen">the site ien as its stored in Claims</param>
        /// <param name="ipAddress"></param>
        /// <param name="progress">optional; provide a progress so we can report the status</param>
        /// <returns></returns>
        public async Task<SynchronizationResult> SynchronizeToSite(string vistaSiteIen,
            string ipAddress, IProgress<ProgressDialog> progress = null)
        {
            _progress = progress;

            var siteIens = new List<string>() {vistaSiteIen};

            // get the most recent claims profile
            await UpdateProgress("Refreshing Claims user profile", 10);
            var claimsUser = _vistaUserService.GetVistaUser();

            // fetch a list of forbidden keys
            await UpdateProgress("Refreshing forbidden keys", 20);
            var forbiddenKeys = GetForbiddenKeys();

            // fetch a list of sites that the user can sync 
            await UpdateProgress("Verifying permission to sync", 30);
            var syncableSites = CanUserSyncToSites(siteIens);

            // if we don't have any sites to sync then return
            if (syncableSites.Count < 1)
            {
                return null;
            }

            var vistaSite = syncableSites.First().Value;

            // is the primary menu on claims empty?
            if (string.IsNullOrEmpty(claimsUser.PrimaryMenu))
            {
                // TODO this should be checked on first login
            }

            // is the primary menu authorized for syncing to remote vista instances?
            if (!IsAuthorizedPrimaryMenu(claimsUser))
            {
                // TODO this should be checked on first login
            }

            // perform synchronization
            await UpdateProgress("Synchronizing your VistA account", 80);
            var synchronizationResult = SynchronizeToSite(claimsUser, vistaSite, forbiddenKeys, ipAddress);

            return synchronizationResult;
        }

        private bool IsAuthorizedPrimaryMenu(VistaUser claimsUser)
        {
            using (var rpcClient = GetRpcClient())
            {
                var ddrFindOne = new DdrFindOne(file: "100100.0341", lookupValue: claimsUser.PrimaryMenu);
            
                rpcClient.CallAndDisconnect(ddrFindOne);

                if (ddrFindOne.GetResult().Equals("0"))
                {
                    return false;
                }

                return true;
            }
        }

        private SynchronizationResult SynchronizeToSite(VistaUser claimsUser, VistaSite vistaSite,
            HashSet<string> forbiddenKeys, string ipAddress)
        {
            if (null == claimsUser)
                throw new ArgumentNullException(nameof(claimsUser));
            if (null == vistaSite)
                throw new ArgumentNullException(nameof(vistaSite));
            if (string.IsNullOrEmpty(ipAddress))
                throw new ArgumentNullException(nameof(ipAddress));

            var syncResult = new SynchronizationResult {VistaSite = vistaSite, ClaimsUser = claimsUser};

            using (var rpcClient = GetRpcClient(tryToLogin: false, vistaHostname: vistaSite.VistaHostname,
                vistaPort: Int32.Parse(vistaSite.VistaPort)))
            {
                var duz = "0";

                try
                {
                    // login using bse
                    rpcClient.Connect();
                    var signonSetup = new SignonSetup(BseApplicationPassPhrase, BseToken);
                    rpcClient.Call(signonSetup);

                    // get the user information
                    var getUserInfo = new GetUserInfo();
                    rpcClient.Call(getUserInfo);

                    // throws exception when weird edge case occurs - that is a user is unable to fully signon using bse
                    if (getUserInfo.Response.Contains("User not fully sign-on"))
                    {
                        throw new UserNotSignedOnException(getUserInfo.Response);
                    }

                    var newPerson = getUserInfo.GetNewPerson();
                    duz = newPerson.Ien;

                    syncResult.SequentialLog.Add(new LogEntry()
                    {
                        LogType = LogEntry.LogEntryType.Notice,
                        LogMessage = "Successfully connected to VistA site: " + vistaSite.DisplayName
                    });
                }
                catch (RpcSocketException e)
                {
                    var errorMessage = $"Unable to connect to VistA site {vistaSite.DisplayName}.";
                    syncResult.ErrorsList.Add(errorMessage);

                    syncResult.SequentialLog.Add(new LogEntry()
                    {
                        ErrorOccurred = true,
                        LogType = LogEntry.LogEntryType.Error,
                        LogMessage = errorMessage,
                        LogDetails = e.Message
                    });

                    rpcClient.Disconnect();
                    return syncResult;
                }

                // switch to the vram context
                rpcClient.CreateContext(VistaVramMenuOption);

                // check name using business rules
                var passedBusinessRules = CheckUsingBusinessRules(rpcClient, syncResult, duz);
                if (passedBusinessRules == false)
                {
                    rpcClient.Disconnect();
                    return syncResult;
                }

                // check to see if access code already exists
                var modifiedAccessCode = claimsUser.AccessCode;
                var lastCharacter = modifiedAccessCode[modifiedAccessCode.Length - 1];
                lastCharacter = (char) (lastCharacter - 1);
                modifiedAccessCode = modifiedAccessCode.Remove(modifiedAccessCode.Length - 1) + lastCharacter;

                var accessCodeLister = new DdrLister(file: "200", fields: ".01;2", flags: "I", number: "3", index: "A",
                    @from: modifiedAccessCode);
                rpcClient.Call(accessCodeLister);
                var accessCodes = accessCodeLister.GetAllRows();

                foreach (var accessCodeRow in accessCodes)
                {
                    var personIen = accessCodeRow.Key;

                    var row = accessCodeRow.Value;
                    if (row["2I"].Equals(claimsUser.AccessCode) && !personIen.Equals(duz))
                    {
                        var errorMessage =
                            $"User’s access code already belongs to another user on {vistaSite.DisplayName} site.";
                        syncResult.ErrorsList.Add(errorMessage);

                        rpcClient.Disconnect();
                        return syncResult;
                    }
                }

                // get the user's information on the remote vista system
                const string userFields =
                    ".01;2;11;5;7;8;9;9.2;29;31.3;20.4;20.2;20.3;30;41;200.04;200.19;200.1;200.18;201";
                var userGets = new DdrGets(file: "200", iens: duz + ",", fields: userFields, optionalFlags: "IE");
                rpcClient.Call(userGets);

                var remoteUser = new VistaUser
                {
                    PersonIen = duz,
                    PersonName = userGets.GetFieldValue(".01"),
                    AccessCode = userGets.GetFieldValue("2", false),
                    VerifyCode = userGets.GetFieldValue("11", false),
                    DateOfBirth = userGets.GetFieldValue("5", false),
                    SocialSecurityNumber = userGets.GetFieldValue("9", false),
                    Title = userGets.GetFieldValue("8"),
                    ServiceSection = userGets.GetFieldValue("29"),
                    ScreenEditor = userGets.GetFieldValue("31.3"),
                    ElectronicSignature = userGets.GetFieldValue("20.4", false),
                    ElectronicSignatureTitleBlock = userGets.GetFieldValue("20.3"),
                    ElectronicSignaturePrintedBlock = userGets.GetFieldValue("20.2"),
                    CreateDate = userGets.GetFieldValue("30", false),
                    AllowedToUseSpooler = userGets.GetFieldValue("41"),
                    TimedRead = "900", // setup default value
                    PrimaryMenu = userGets.GetFieldValue("201"),
                    SecondaryMenus = new Dictionary<string, string>(),
                    SecurityKeys = new Dictionary<string, string>(),
                    CprsTabs = new Dictionary<string, CprsTab>()
                };

                // if the claims user timed read is > 900 then use that instead
                if (!String.IsNullOrEmpty(claimsUser.TimedRead))
                {
                    var tryParse = int.TryParse(claimsUser.TimedRead, out var timedRead);
                    if (tryParse && timedRead > 900)
                    {
                        remoteUser.TimedRead = claimsUser.TimedRead;
                    }
                    else
                    {
                        remoteUser.TimedRead = "900";
                    }
                }

                // include the remote user info in the return value
                syncResult.RemoteUser = remoteUser;

                // lookup the primary menu ien on the remote vista
                var primaryMenuFindOne = new DdrFindOne(file: "19", lookupValue: claimsUser.PrimaryMenu, flags: "X");
                rpcClient.Call(primaryMenuFindOne);
                var primaryMenuIen = primaryMenuFindOne.GetResult();

                // fetch all secondary menu options that the user has on remote vista
                remoteUser.SecondaryMenus = GetSecondaryMenus(rpcClient, syncResult);

                // fetch all security keys that the user has on remote vista
                var securityKeysLister = new DdrLister(file: "200.051", iens: "," + duz + ",", fields: ".01;.01I");
                rpcClient.Call(securityKeysLister);
                var allSecurityKeys = securityKeysLister.GetAllRows();

                foreach (var row in allSecurityKeys)
                {
                    var rowMap = row.Value;
                    remoteUser.SecurityKeys.Add(row.Key, rowMap[".01"]);
                }

                // check and if necessary fix corrupt cprs tabs
                var mCode =
                    "n node s node=$o(^VA(200,{0},\"ORD\",\"B\",\"A\")) i (node'=\"\")&&(+node=0) k ^VA(200,{0},\"ORD\") s ^VA(200,{0},\"ORD\",0)=\"^200.010113POI^^0\"";
                mCode = String.Format(mCode, duz);
                var fixCprsTabs = new DdrLister(file: "5", fields: ".01", number: "1", screen: mCode);
                rpcClient.Call(fixCprsTabs);

                /*
                Example of CORRUPT nodes...
                ---------------------------------------------
                ^VA(200,7287507,8910,"B",359,1)=""
                ^VA(200,7287507,"ORD",0)="^200.010113POI^1^1"
                ^VA(200,7287507,"ORD",1,0)="COR^JUL 07, 2015^AUG 07, 2015"
                ^VA(200,7287507,"ORD","B","COR",1)=""
                ^VA(200,7287507,"SPL")=""
                */

                // fetch all cprs tabs that the user has on remote vista
                var cprsTabsLister = new DdrLister(file: "200.010113", iens: "," + duz + ",",
                    fields: ".01I;.02I;.03I;.01;.02;.03;", flags: "P");
                rpcClient.Call(cprsTabsLister);
                var allCprsTabs = cprsTabsLister.GetAllRows();

                foreach (var row in allCprsTabs)
                {
                    var rowMap = row.Value;
                    var cprsTab = new CprsTab
                    {
                        CprsTabIen = row.Key,
                        CprsTabName = rowMap[".01I"],
                        EffectiveDate = rowMap[".02I"],
                        ExpirationDate = rowMap[".03I"]
                    };
                    remoteUser.CprsTabs.Add(row.Key, cprsTab);
                }

                // lookup the user's title ien on the remote vista 
                var titleFindOne = new DdrFindOne(file: "3.1", lookupValue: claimsUser.Title, flags: "X");
                rpcClient.Call(titleFindOne);
                var titleIen = titleFindOne.GetResult();

                // lookup the user's service/section ien on the remote vista
                var serviceSectionIen = FindOrAddServiceSection(rpcClient, syncResult);

                #region comparing claims versus remote vista

                // step 5 - compare and update old data on remote vista system
                var fileManDataArray = new FileManDataArray();
                DdrFiler ddrFiler;

                // setup fields that will be PUSHED
                fileManDataArray.Add(FileNumber: "200", Iens: remoteUser.PersonIen + ",", FieldNumber: "2", Value: "@");
                // access code
                fileManDataArray.Add(FileNumber: "200", Iens: remoteUser.PersonIen + ",", FieldNumber: "7", Value: "@");
                // disuser
                fileManDataArray.Add(FileNumber: "200", Iens: remoteUser.PersonIen + ",", FieldNumber: "9.2",
                    Value: "T+30d");
                // termination date
                fileManDataArray.Add(FileNumber: "200", Iens: remoteUser.PersonIen + ",", FieldNumber: "9.21",
                    Value: "YES");
                // delete all mail access
                fileManDataArray.Add(FileNumber: "200", Iens: remoteUser.PersonIen + ",", FieldNumber: "9.22",
                    Value: "YES");
                // delete keys at termination
                fileManDataArray.Add(FileNumber: "200", Iens: remoteUser.PersonIen + ",", FieldNumber: "101.01",
                    Value: "NO");
                // restrict patient selection
                fileManDataArray.Add(FileNumber: "200", Iens: remoteUser.PersonIen + ",", FieldNumber: "200.04",
                    Value: "ALLOWED"); // multiple signon
                fileManDataArray.Add(FileNumber: "200", Iens: remoteUser.PersonIen + ",", FieldNumber: "200.19",
                    Value: "3");
                // multiple signon limit
                fileManDataArray.Add(FileNumber: "200", Iens: remoteUser.PersonIen + ",", FieldNumber: "200.1",
                    Value: remoteUser.TimedRead); // timed read
                fileManDataArray.Add(FileNumber: "200", Iens: remoteUser.PersonIen + ",", FieldNumber: "200.18",
                    Value: "No");
                // auto sign on

                if (!String.IsNullOrEmpty(titleIen) && !titleIen.Equals("0"))
                {
                    fileManDataArray.Add(FileNumber: "200", Iens: remoteUser.PersonIen + ",", FieldNumber: "8",
                        Value: claimsUser.Title); // title
                }
                else
                {
                    var warningMessage =
                        $"Could not synchronize {claimsUser.Title} for TITLE field in file 3.1. Please contact your IT department.";
                    syncResult.WarningsList.Add(warningMessage);

                    syncResult.SequentialLog.Add(new LogEntry()
                    {
                        LogType = LogEntry.LogEntryType.Warning,
                        LogMessage = warningMessage
                    });
                }

                if (!String.IsNullOrEmpty(serviceSectionIen))
                    fileManDataArray.Add(FileNumber: "200", Iens: remoteUser.PersonIen + ",", FieldNumber: "29",
                        Value: claimsUser.ServiceSection); // service/section

                if (!claimsUser.ScreenEditor.Equals(remoteUser.ScreenEditor))
                    fileManDataArray.Add(FileNumber: "200", Iens: remoteUser.PersonIen + ",", FieldNumber: "31.3",
                        Value: claimsUser.ScreenEditor); // screen editor
                if (!claimsUser.ElectronicSignaturePrintedBlock.Equals(remoteUser.ElectronicSignaturePrintedBlock))
                    fileManDataArray.Add(FileNumber: "200", Iens: remoteUser.PersonIen + ",", FieldNumber: "20.2",
                        Value: claimsUser.ElectronicSignaturePrintedBlock); // signature printed block
                if (!claimsUser.ElectronicSignatureTitleBlock.Equals(remoteUser.ElectronicSignatureTitleBlock))
                    fileManDataArray.Add(FileNumber: "200", Iens: remoteUser.PersonIen + ",", FieldNumber: "20.3",
                        Value: claimsUser.ElectronicSignatureTitleBlock); // signature title block
                if (!claimsUser.AllowedToUseSpooler.Equals(remoteUser.AllowedToUseSpooler))
                    fileManDataArray.Add(FileNumber: "200", Iens: remoteUser.PersonIen + ",", FieldNumber: "41",
                        Value: claimsUser.AllowedToUseSpooler); // allowed to use spooler

                try
                {
                    ddrFiler = new DdrFiler(mode: DdrFiler.DdrMode.UPDATE, flags: "E", fdaRoot: fileManDataArray);
                    rpcClient.Call(ddrFiler); // update the new person file

                    syncResult.SequentialLog.Add(new LogEntry()
                    {
                        LogType = LogEntry.LogEntryType.Notice,
                        LogMessage = "Successfully updated your personal information on the remote VistA system"
                    });
                }
                catch (RpcSocketException e)
                {
                    syncResult.ErrorsList.Add(e.Message);

                    syncResult.SequentialLog.Add(new LogEntry()
                    {
                        ErrorOccurred = true,
                        LogType = LogEntry.LogEntryType.Error,
                        LogMessage = "Failed to update your personal information in the remote VistA system",
                        LogDetails = e.Message
                    });
                }

                var primaryMenuChanged = false;

                // now file the remaining fields using the internal values
                fileManDataArray = new FileManDataArray();
                var fdaPrimaryMenu = false;
                var fdaAvCodes = false;

                if (!String.IsNullOrEmpty(primaryMenuIen) && !claimsUser.PrimaryMenu.Equals(remoteUser.PrimaryMenu))
                {
                    fileManDataArray.Add(FileNumber: "200", Iens: remoteUser.PersonIen + ",", FieldNumber: "201",
                        Value: primaryMenuIen); // primary menu
                    primaryMenuChanged = true;
                    fdaPrimaryMenu = true;
                }
                else if (String.IsNullOrEmpty(primaryMenuIen) || primaryMenuIen.Equals("0"))
                {
                    var errorMessage =
                        $"Primary Menu \"{claimsUser.PrimaryMenu}\" was not found on {vistaSite.DisplayName} therefore it cannot be synchronized.";
                    syncResult.ErrorsList.Add(errorMessage);

                    syncResult.SequentialLog.Add(new LogEntry()
                    {
                        ErrorOccurred = true,
                        LogType = LogEntry.LogEntryType.Error,
                        LogMessage = errorMessage
                    });
                }

                if (!claimsUser.ElectronicSignature.Equals(remoteUser.ElectronicSignature))
                    fileManDataArray.Add(FileNumber: "200", Iens: remoteUser.PersonIen + ",", FieldNumber: "20.4",
                        Value: claimsUser.ElectronicSignature); // electronic signature code

                // if user has the security key DVBA VRAM AV PUSH then you are allowed to sync a/v pair
                if (claimsUser.SecurityKeys.ContainsValue("DVBA VRAM AV PUSH"))
                {
                    fdaAvCodes = true;
                    fileManDataArray.Add(FileNumber: "200", Iens: remoteUser.PersonIen + ",", FieldNumber: "2",
                        Value: claimsUser.AccessCode); // access code

                    fileManDataArray.Add(FileNumber: "200", Iens: remoteUser.PersonIen + ",", FieldNumber: "11",
                        Value: claimsUser.VerifyCode); // verify code

                    fileManDataArray.Add(FileNumber: "200", Iens: remoteUser.PersonIen + ",", FieldNumber: "11.2",
                        Value: MumpsDateTime.Now()); // $h of verify code last changed
                }
                else
                {
                    var warningMessage = "Access and verify codes were NOT synchronized due to a missing security key.";
                    syncResult.WarningsList.Add(warningMessage);

                    syncResult.SequentialLog.Add(new LogEntry()
                    {
                        LogType = LogEntry.LogEntryType.Warning,
                        LogMessage = warningMessage
                    });
                }

                try
                {
                    ddrFiler = new DdrFiler(mode: DdrFiler.DdrMode.UPDATE, flags: "", fdaRoot: fileManDataArray);
                    rpcClient.Call(ddrFiler);

                    var logEntry = new LogEntry() {LogType = LogEntry.LogEntryType.Notice};
                    if (fdaPrimaryMenu && fdaAvCodes)
                        logEntry.LogMessage = "Successfully updated your primary menu and A/V code pair.";
                    if (!fdaPrimaryMenu && fdaAvCodes)
                        logEntry.LogMessage = "Successfully updated your A/V code pair.";
                    if (fdaPrimaryMenu && !fdaAvCodes)
                        logEntry.LogMessage = "Successfully updated your primary menu.";

                    syncResult.SequentialLog.Add(logEntry);
                }
                catch (RpcSocketException e)
                {
                    var logEntry = new LogEntry()
                    {
                        LogType = LogEntry.LogEntryType.Error,
                        ErrorOccurred = true,
                        LogDetails = e.Message
                    };
                    if (fdaPrimaryMenu && fdaAvCodes)
                        logEntry.LogMessage = "Failed to update your primary menu and A/V code pair.";
                    if (!fdaPrimaryMenu && fdaAvCodes)
                        logEntry.LogMessage = "Failed to update your A/V code pair.";
                    if (fdaPrimaryMenu && !fdaAvCodes)
                        logEntry.LogMessage = "Failed to update your primary menu.";

                    syncResult.SequentialLog.Add(logEntry);

                    syncResult.ErrorsList.Add(logEntry.LogMessage);
                }

                var secondaryMenusChanged = UpdateSecondaryMenus(rpcClient, syncResult);
                UpdateSecurityKeys(rpcClient, syncResult, forbiddenKeys);
                UpdateCprsTabs(rpcClient, syncResult);

                #endregion

                // if any of the menus changed call the rpc XU REBUILD MENU TREE
                if (primaryMenuChanged || secondaryMenusChanged)
                {
                    var menuRebuild = new RebuildMenuTree();
                    try
                    {
                        rpcClient.Call(menuRebuild);
                        if (menuRebuild.IsRebuildMenuTreeQueued() == false)
                        {
                            var errorMessage =
                                $"Menu tree rebuild for user failed. Please attempt to synchronize to {vistaSite.DisplayName} again. If it continues to fail, contact your IT support person.";

                            syncResult.ErrorsList.Add(errorMessage);

                            syncResult.SequentialLog.Add(new LogEntry()
                            {
                                ErrorOccurred = true,
                                LogType = LogEntry.LogEntryType.Error,
                                LogMessage = errorMessage
                            });
                        }
                    }
                    catch (RpcSocketException e)
                    {
                        var errorMessage =
                            $"Menu tree rebuild for user failed. Please attempt to synchronize to {vistaSite.DisplayName} again. If it continues to fail, contact your IT support person. Details: {e.Message}";
                        syncResult.ErrorsList.Add(errorMessage);

                        syncResult.SequentialLog.Add(new LogEntry()
                        {
                            ErrorOccurred = true,
                            LogType = LogEntry.LogEntryType.Error,
                            LogMessage = errorMessage
                        });
                    }
                }

                // fetch a handle to a token that will sign-on a new process.
                var tokenGet = new TokenGet();
                rpcClient.Call(tokenGet);
                syncResult.SignOnToken = tokenGet.Token;

                if (ipAddress != "::1")
                {
                    var fixTokenCode = "i $d(^XTMP(\"{0}\"))>0 s ^XTMP(\"{0}\",\"D3\")=\"{1}\"";
                    fixTokenCode = string.Format(fixTokenCode, tokenGet.Token.Substring(2), ipAddress);

                    var fixToken = new DdrLister(file: "5", fields: ".01", number: "1", screen: fixTokenCode);
                    rpcClient.Call(fixToken);
                }

                rpcClient.Disconnect();

                syncResult.SuccessfullySynchronized = true;

                return syncResult;
            }
        }

        private Dictionary<string, string> GetSecondaryMenus(IRpcClient rpcClient, SynchronizationResult syncResult)
        {
            var secondaryMenusLister = new DdrLister(file: "200.03", iens: "," + syncResult.RemoteUser.PersonIen + ",",
                fields: ".01;.01I");
            rpcClient.Call(secondaryMenusLister);
            var allSecondaryMenus = secondaryMenusLister.GetAllRows();

            var result = new Dictionary<string, string>();

            foreach (var row in allSecondaryMenus)
            {
                var rowMap = row.Value;
                result.Add(row.Key, rowMap[".01"]);
            }

            return result;
        }

        private string FindOrAddServiceSection(IRpcClient rpcClient, SynchronizationResult syncResult)
        {
            var serviceFindOne = new DdrFindOne(file: "49", lookupValue: syncResult.ClaimsUser.ServiceSection,
                flags: "X");
            rpcClient.Call(serviceFindOne);
            var serviceIen = serviceFindOne.GetResult();

            if (!serviceIen.Equals("0"))
                return serviceIen;

            var fdaRoot = new FileManDataArray();
            fdaRoot.Add(FileNumber: "49", Iens: "+1,", FieldNumber: ".01", Value: syncResult.ClaimsUser.ServiceSection);

            try
            {
                var ddrFiler = new DdrFiler(mode: DdrFiler.DdrMode.ADD, flags: "E", fdaRoot: fdaRoot);
                rpcClient.Call(ddrFiler);
            }
            catch (RpcSocketException)
            {
                string errorMessage =
                    $"Service/Section not added to file #49 at site {syncResult.VistaSite.DisplayName}. Please contact your IT department.";
                syncResult.ErrorsList.Add(errorMessage);

                syncResult.SequentialLog.Add(new LogEntry()
                {
                    ErrorOccurred = true,
                    LogType = LogEntry.LogEntryType.Error,
                    LogMessage = errorMessage
                });

                return null;
            }

            rpcClient.Call(serviceFindOne);
            serviceIen = serviceFindOne.GetResult();

            return serviceIen;
        }

        /// <summary>
        /// this implementation is sourced from clsUserInfo.pas:620 --> function TUsrInfo.SyncNameDOB(Brkr: TVramRPCBroker): Boolean;
        /// </summary>
        /// <param name="rpcClient"></param>
        /// <param name="syncResult"></param>
        /// <param name="remoteUserDuz"></param>
        /// <returns></returns>
        private bool CheckUsingBusinessRules(IRpcClient rpcClient, SynchronizationResult syncResult,
            string remoteUserDuz)
        {
            var iens = remoteUserDuz + ",";
            var ddrGets = new DdrGets(file: "200", iens: iens, fields: ".01;5;30;201", optionalFlags: "IE");
            rpcClient.Call(ddrGets);

            var remoteFullName = ddrGets.GetFieldValue(".01", false);
            var remoteDateOfBirth = ddrGets.GetFieldValue("5", false);
            var remoteCreateDate = ddrGets.GetFieldValue("30", false);
            var remotePrimaryMenu = ddrGets.GetFieldValue("201");

            var claimsLastFirstName = syncResult.ClaimsUser.PersonName.Split(" ")[0];
            // remove the middle name or initial
            var claimsDateOfBirth = syncResult.ClaimsUser.DateOfBirth;

            var fileManDate = FileManDateTime.ToFileManTimeStamp(DateTime.Now).Split(".")[0];
            var isNewEntry = (String.IsNullOrEmpty(remotePrimaryMenu) &&
                              (String.IsNullOrEmpty(remoteCreateDate) || remoteCreateDate.Equals(fileManDate)));

            if (isNewEntry)
            {
                if (!syncResult.VistaSite.IgnoreSimilarName)
                {
                    var modifiedLastFirstName = claimsLastFirstName;
                    var lastCharacter = modifiedLastFirstName[modifiedLastFirstName.Length - 1];
                    lastCharacter = (char) (lastCharacter - 1);
                    modifiedLastFirstName = modifiedLastFirstName.Remove(modifiedLastFirstName.Length - 1) +
                                            lastCharacter;

                    var ddrLister = new DdrLister(file: "200", fields: ".01", @from: modifiedLastFirstName, number: "3",
                        index: "B");
                    rpcClient.Call(ddrLister);

                    var allRows = ddrLister.GetAllRows();
                    if (allRows.Count >= 2)
                    {
                        var firstResultName = allRows[allRows.Keys.ElementAt(0)][".01"];
                        var secondResultName = allRows[allRows.Keys.ElementAt(1)][".01"];

                        if (firstResultName.Substring(0, claimsLastFirstName.Length).Equals(claimsLastFirstName))
                        {
                            if (secondResultName.Substring(0, claimsLastFirstName.Length).Equals(claimsLastFirstName))
                            {
                                var errorMessage =
                                    $"An account was created for you on the VistA system {syncResult.VistaSite.DisplayName} but you need to alert your IT department to verify this account.";

                                syncResult.SequentialLog.Add(new LogEntry()
                                {
                                    ErrorOccurred = true,
                                    LogType = LogEntry.LogEntryType.Warning,
                                    LogMessage = errorMessage
                                });

                                syncResult.WarningsList.Add(errorMessage);
                                return true;
                            }
                        }
                    }
                }
            }
            else
            {
                var fdaRoot = new FileManDataArray();

                var remoteLastFirstName = remoteFullName.Split(" ")[0];
                if (claimsLastFirstName.Equals(remoteLastFirstName))
                {
                    if (!syncResult.ClaimsUser.PersonName.Equals(remoteFullName))
                    {
                        fdaRoot.Add(FileNumber: "200", Iens: iens, FieldNumber: ".01",
                            Value: syncResult.ClaimsUser.PersonName);
                    }

                    if ((!claimsDateOfBirth.Equals("")) && (!claimsDateOfBirth.Equals(remoteDateOfBirth)))
                    {
                        fdaRoot.Add(FileNumber: "200", Iens: iens, FieldNumber: "5",
                            Value: syncResult.ClaimsUser.DateOfBirth);
                    }
                }
                else
                {
                    if (syncResult.VistaSite.OverwriteName)
                    {
                        fdaRoot.Add(FileNumber: "200", Iens: iens, FieldNumber: ".01",
                            Value: syncResult.ClaimsUser.PersonName);
                    }
                    else
                    {
                        var errorMessage =
                            $"Cannot create an account for you on VistA system {syncResult.VistaSite.DisplayName} because of an accounting matching issue. Please contact your IT department. They will have to remediate the issue before an account can be created at this site.";
                        syncResult.ErrorsList.Add(errorMessage);

                        syncResult.SequentialLog.Add(new LogEntry()
                        {
                            ErrorOccurred = true,
                            LogType = LogEntry.LogEntryType.Error,
                            LogMessage = errorMessage
                        });

                        return false;
                    }
                }

                if (fdaRoot.Count > 0)
                {
                    try
                    {
                        var ddrFiler = new DdrFiler(mode: DdrFiler.DdrMode.UPDATE, fdaRoot: fdaRoot, flags: "");
                        rpcClient.Call(ddrFiler);
                    }
                    catch (RpcSocketException e)
                    {
                        var errorMessage =
                            "Unknown error occurred while attempting to update NAME and/or DOB. Cannot continue at this site until problems are resolved.";
                        syncResult.ErrorsList.Add(errorMessage);

                        syncResult.SequentialLog.Add(new LogEntry()
                        {
                            ErrorOccurred = true,
                            LogType = LogEntry.LogEntryType.Error,
                            LogMessage = errorMessage,
                            LogDetails = e.Message
                        });

                        return false;
                    }
                }
            }

            return true;
        }

        private bool UpdateSecondaryMenus(IRpcClient rpcClient, SynchronizationResult synchronizationResult)
        {
            var menusChanged = false;

            var claimsUserSecondaryMenus = synchronizationResult.ClaimsUser.SecondaryMenus;
            var remoteUserSecondaryMenus = synchronizationResult.RemoteUser.SecondaryMenus;

            FileManDataArray fdaRoot;

            // duplicate menu check and removal
            var duplicateMenus =
                remoteUserSecondaryMenus.GroupBy(i => i.Value).Where(g => g.Count() > 1).Select(g => g);
            foreach (var duplicateMenu in duplicateMenus)
            {
                var menuOptions = duplicateMenu.ToList();
                var enumerator = menuOptions.GetEnumerator();

                for (int i = 0; i < menuOptions.Count() - 1; i++)
                {
                    enumerator.MoveNext();
                    var menuOption = enumerator.Current;
                    var remotePersonIen = synchronizationResult.RemoteUser.PersonIen;
                    var entryIen = menuOption.Key;

                    fdaRoot = new FileManDataArray();

                    var iens = $"{entryIen},{remotePersonIen},";
                    fdaRoot.Add(FileNumber: "200.03", Iens: iens, FieldNumber: ".01", Value: "@");

                    try
                    {
                        if (fdaRoot.Count > 0)
                        {
                            var ddrFiler = new DdrFiler(mode: DdrFiler.DdrMode.UPDATE, fdaRoot: fdaRoot, flags: "E");
                            rpcClient.Call(ddrFiler);

                            menusChanged = true;

                            synchronizationResult.SequentialLog.Add(new LogEntry()
                            {
                                ErrorOccurred = false,
                                LogType = LogEntry.LogEntryType.Notice,
                                LogMessage = "Successfully removed duplicate secondary menu option: " + menuOption.Value
                            });
                        }
                    }
                    catch (RpcSocketException)
                    {
                        var errorMessage =
                            $"Unsuccessful attempt to remove duplicate {menuOption.Value} secondary option at site {synchronizationResult.VistaSite.DisplayName}.";
                        synchronizationResult.ErrorsList.Add(errorMessage);

                        synchronizationResult.SequentialLog.Add(new LogEntry()
                        {
                            LogType = LogEntry.LogEntryType.Warning,
                            LogMessage = errorMessage
                        });
                    }
                }
            }

            // add secondary menus
            foreach (var menuOption in claimsUserSecondaryMenus)
            {
                if (!remoteUserSecondaryMenus.ContainsValue(menuOption.Value))
                {
                    fdaRoot = new FileManDataArray();

                    var ddrFindOne = new DdrFindOne(file: "19", lookupValue: menuOption.Value, flags: "X");
                    rpcClient.Call(ddrFindOne);
                    var menuIen = ddrFindOne.GetResult();

                    if (!String.IsNullOrEmpty(menuIen) && menuIen != "0")
                    {
                        var iens = $"+1,{synchronizationResult.RemoteUser.PersonIen},";
                        fdaRoot.Add(FileNumber: "200.03", Iens: iens, FieldNumber: ".01", Value: menuIen);
                    }
                    else
                    {
                        var message =
                            $"Unable to add {menuOption.Value} as secondary option because it is not found in file #19 on {synchronizationResult.VistaSite.DisplayName} site.";
                        synchronizationResult.WarningsList.Add(message);

                        synchronizationResult.SequentialLog.Add(new LogEntry()
                        {
                            LogType = LogEntry.LogEntryType.Warning,
                            LogMessage = message
                        });
                    }

                    try
                    {
                        if (fdaRoot.Count > 0)
                        {
                            var ddrFiler = new DdrFiler(mode: DdrFiler.DdrMode.ADD, fdaRoot: fdaRoot, flags: "");
                            rpcClient.Call(ddrFiler);

                            menusChanged = true;

                            synchronizationResult.SequentialLog.Add(new LogEntry()
                            {
                                LogType = LogEntry.LogEntryType.Notice,
                                LogMessage = "Successfully added secondary menu option: " + menuOption.Value
                            });
                        }
                    }
                    catch (RpcSocketException)
                    {
                        var errorMessage = $"Unsuccessful attempt to add {menuOption.Value} as a secondary option.";
                        synchronizationResult.ErrorsList.Add(errorMessage);

                        synchronizationResult.SequentialLog.Add(new LogEntry()
                        {
                            ErrorOccurred = true,
                            LogType = LogEntry.LogEntryType.Error,
                            LogMessage = errorMessage
                        });
                    }
                }
            }

            fdaRoot = new FileManDataArray();

            // remove secondary menus
            foreach (var menuOption in remoteUserSecondaryMenus)
            {
                if (!claimsUserSecondaryMenus.ContainsValue(menuOption.Value))
                {
                    var remotePersonIen = synchronizationResult.RemoteUser.PersonIen;
                    var entryIen = menuOption.Key;

                    var iens = $"{entryIen},{remotePersonIen},";
                    fdaRoot.Add(FileNumber: "200.03", Iens: iens, FieldNumber: ".01", Value: "@");
                }

                try
                {
                    if (fdaRoot.Count > 0)
                    {
                        var ddrFiler = new DdrFiler(mode: DdrFiler.DdrMode.UPDATE, fdaRoot: fdaRoot, flags: "E");
                        rpcClient.Call(ddrFiler);

                        menusChanged = true;

                        synchronizationResult.SequentialLog.Add(new LogEntry()
                        {
                            LogType = LogEntry.LogEntryType.Notice,
                            LogMessage = "Successfully removed secondary menu option: " + menuOption.Value
                        });
                    }
                }
                catch (RpcSocketException)
                {
                    var errorMessage =
                        $"Unsuccessful attempt to remove {menuOption.Value} secondary option at site {synchronizationResult.VistaSite.DisplayName}.";
                    synchronizationResult.ErrorsList.Add(errorMessage);

                    synchronizationResult.SequentialLog.Add(new LogEntry()
                    {
                        LogType = LogEntry.LogEntryType.Warning,
                        LogMessage = errorMessage
                    });
                }
            }

            return menusChanged;
        }

        private void UpdateSecurityKeys(IRpcClient rpcClient, SynchronizationResult synchronizationResult,
            HashSet<string> forbiddenKeys)
        {
            var claimsSecurityKeys = synchronizationResult.ClaimsUser.SecurityKeys;
            var remoteSecurityKeys = synchronizationResult.RemoteUser.SecurityKeys;

            FileManDataArray fdaRoot;

            // add security keys
            foreach (var claimSecurityKey in claimsSecurityKeys)
            {
                if (forbiddenKeys.Contains(claimSecurityKey.Value))
                {
                    string message = "Skipping forbidden key: " + claimSecurityKey.Value;
                    synchronizationResult.WarningsList.Add(message);

                    synchronizationResult.SequentialLog.Add(new LogEntry()
                    {
                        LogType = LogEntry.LogEntryType.Notice,
                        LogMessage = message
                    });

                    continue;
                }

                if (!remoteSecurityKeys.ContainsValue(claimSecurityKey.Value))
                {
                    var ddrFindOne = new DdrFindOne(file: "19.1", lookupValue: claimSecurityKey.Value, flags: "X");
                    rpcClient.Call(ddrFindOne);
                    var securityKeyIen = ddrFindOne.GetResult();

                    if (!String.IsNullOrEmpty(securityKeyIen) && securityKeyIen != "0")
                    {
                        fdaRoot = new FileManDataArray();

                        var iens = $"+1,{synchronizationResult.RemoteUser.PersonIen},";
                        fdaRoot.Add(FileNumber: "200.051", Iens: iens, FieldNumber: ".01", Value: securityKeyIen);
                        fdaRoot.Add(FileNumber: "200.051", Iens: iens, FieldNumber: "1", Value: ".5");

                        try
                        {
                            var ienRoot = new List<string> {securityKeyIen};
                            var ddrFiler = new DdrFiler(mode: DdrFiler.DdrMode.ADD, fdaRoot: fdaRoot, flags: "",
                                ienRoot: ienRoot);
                            rpcClient.Call(ddrFiler);

                            synchronizationResult.SequentialLog.Add(new LogEntry()
                            {
                                LogType = LogEntry.LogEntryType.Notice,
                                LogMessage = "Succesfully added the security key: " + claimSecurityKey.Value
                            });
                        }
                        catch (RpcSocketException)
                        {
                            var errorMessage = $"Unsuccessful attempt to add {claimSecurityKey.Value} key.";
                            synchronizationResult.ErrorsList.Add(errorMessage);

                            synchronizationResult.SequentialLog.Add(new LogEntry()
                            {
                                LogType = LogEntry.LogEntryType.Warning,
                                LogMessage = errorMessage
                            });
                        }
                    }
                    else
                    {
                        var warningMessage =
                            $"Unable to add {claimSecurityKey.Value} as security key because it is not found in file #19.1 on {synchronizationResult.VistaSite.DisplayName} site.";
                        synchronizationResult.WarningsList.Add(warningMessage);

                        synchronizationResult.SequentialLog.Add(new LogEntry()
                        {
                            LogType = LogEntry.LogEntryType.Warning,
                            LogMessage = warningMessage
                        });
                    }
                }
            }

            // remove security keys
            foreach (var remoteSecurityKey in remoteSecurityKeys)
            {
                if (!claimsSecurityKeys.ContainsValue(remoteSecurityKey.Value) ||
                    forbiddenKeys.Contains(remoteSecurityKey.Value))
                {
                    var remotePersonIen = synchronizationResult.RemoteUser.PersonIen;
                    var securityKeyIen = remoteSecurityKey.Key;

                    if (!String.IsNullOrEmpty(securityKeyIen) && securityKeyIen != "0")
                    {
                        fdaRoot = new FileManDataArray();
                        var iens = $"{securityKeyIen},{remotePersonIen},";
                        fdaRoot.Add(FileNumber: "200.051", Iens: iens, FieldNumber: ".01", Value: "@");

                        try
                        {
                            if (fdaRoot.Count > 0)
                            {
                                var ddrFiler = new DdrFiler(mode: DdrFiler.DdrMode.UPDATE, fdaRoot: fdaRoot,
                                    flags: "E");
                                rpcClient.Call(ddrFiler);

                                synchronizationResult.SequentialLog.Add(new LogEntry()
                                {
                                    LogType = LogEntry.LogEntryType.Notice,
                                    LogMessage = "Successfully removed security key: " + remoteSecurityKey.Value
                                });
                            }
                        }
                        catch (RpcSocketException)
                        {
                            var errorMessage =
                                $"Unsuccessful attempt to remove {remoteSecurityKey.Value} key at site {synchronizationResult.VistaSite.DisplayName}.";
                            synchronizationResult.ErrorsList.Add(errorMessage);

                            synchronizationResult.SequentialLog.Add(new LogEntry()
                            {
                                LogType = LogEntry.LogEntryType.Warning,
                                LogMessage = errorMessage
                            });
                        }
                    }
                }
            }
        }

        private void UpdateCprsTabs(IRpcClient rpcClient, SynchronizationResult synchronizationResult)
        {
            var claimsCprsTabs = synchronizationResult.ClaimsUser.CprsTabs;
            var remoteCprsTabs = synchronizationResult.RemoteUser.CprsTabs;

            FileManDataArray fdaRoot;

            var claimsHashSet = new HashSet<string>(claimsCprsTabs.Select(x => x.Value.CprsTabName).ToList());
            var remoteHashSet = new HashSet<string>(remoteCprsTabs.Select(x => x.Value.CprsTabName).ToList());
            var remotePersonIen = synchronizationResult.RemoteUser.PersonIen;

            // just delete all the cprs tabs
            foreach (var remoteCprsTab in remoteCprsTabs)
            {
                var cprsTabIen = remoteCprsTab.Key;

                if (!String.IsNullOrEmpty(cprsTabIen) && cprsTabIen != "0")
                {
                    fdaRoot = new FileManDataArray();
                    var iens = $"{cprsTabIen},{remotePersonIen},";
                    fdaRoot.Add(FileNumber: "200.010113", Iens: iens, FieldNumber: ".01", Value: "@");

                    try
                    {
                        if (fdaRoot.Count > 0)
                        {
                            var ddrFiler = new DdrFiler(mode: DdrFiler.DdrMode.UPDATE, fdaRoot: fdaRoot, flags: "E");
                            rpcClient.Call(ddrFiler);

                            var cprsTabLog = "Unknown";
                            if (remoteCprsTab.Value.CprsTabName.Equals("1"))
                                cprsTabLog = "COR";
                            else if (remoteCprsTab.Value.CprsTabName.Equals("2"))
                                cprsTabLog = "RPT";

                            synchronizationResult.SequentialLog.Add(new LogEntry()
                            {
                                LogType = LogEntry.LogEntryType.Notice,
                                LogMessage = "Successfully removed CPRS tab: " + cprsTabLog
                            });
                        }
                    }
                    catch (RpcSocketException)
                    {
                        var cprsTabLog = "Unknown";
                        if (remoteCprsTab.Value.CprsTabName.Equals("1"))
                            cprsTabLog = "COR";
                        else if (remoteCprsTab.Value.CprsTabName.Equals("2"))
                            cprsTabLog = "RPT";

                        string errorMessage = "Failed to remove CPRS tab: " + cprsTabLog;
                        synchronizationResult.ErrorsList.Add(errorMessage);

                        synchronizationResult.SequentialLog.Add(new LogEntry()
                        {
                            ErrorOccurred = true,
                            LogType = LogEntry.LogEntryType.Error,
                            LogMessage = errorMessage
                        });
                    }
                }
            }

            // add cprs tabs
            foreach (var claimsCprsTab in claimsCprsTabs)
            {
                var cprsTabIen = claimsCprsTab.Key;
                var cprsTab = claimsCprsTab.Value;

                fdaRoot = new FileManDataArray();

                var iens = $"+1,{remotePersonIen},";
                fdaRoot.Add(FileNumber: "200.010113", Iens: iens, FieldNumber: ".01", Value: cprsTab.CprsTabName);
                fdaRoot.Add(FileNumber: "200.010113", Iens: iens, FieldNumber: ".02", Value: cprsTab.EffectiveDate);
                fdaRoot.Add(FileNumber: "200.010113", Iens: iens, FieldNumber: ".03", Value: cprsTab.ExpirationDate);

                try
                {
                    if (fdaRoot.Count > 0)
                    {
                        var ddrFiler = new DdrFiler(mode: DdrFiler.DdrMode.ADD, fdaRoot: fdaRoot, flags: "");
                        rpcClient.Call(ddrFiler);

                        var cprsTabLog = "Unknown";
                        if (cprsTab.CprsTabName.Equals("1"))
                            cprsTabLog = "COR";
                        else if (cprsTab.CprsTabName.Equals("2"))
                            cprsTabLog = "RPT";

                        synchronizationResult.SequentialLog.Add(new LogEntry()
                        {
                            LogType = LogEntry.LogEntryType.Notice,
                            LogMessage = "Successfully added CPRS tab: " + cprsTabLog
                        });
                    }
                }
                catch (RpcSocketException)
                {
                    var cprsTabLog = "Unknown";
                    if (cprsTab.CprsTabName.Equals("1"))
                        cprsTabLog = "COR";
                    else if (cprsTab.CprsTabName.Equals("2"))
                        cprsTabLog = "RPT";

                    string errorMessage = "Failed to add CPRS tab: " + cprsTabLog;
                    synchronizationResult.ErrorsList.Add(errorMessage);

                    synchronizationResult.SequentialLog.Add(new LogEntry()
                    {
                        ErrorOccurred = true,
                        LogType = LogEntry.LogEntryType.Error,
                        LogMessage = errorMessage
                    });
                }
            }
        }

        /// <summary>
        /// Get a dictionary of all the VistA sites a user has permission to synchronize with
        /// </summary>
        /// <returns>a dictionary of vista sites; key = vistaSiteIen, value = VistaSite</returns>
        public Dictionary<string, VistaSite> GetVistaSites()
        {
            using (var rpcClient = GetRpcClient())
            {
                var vistaSites = new Dictionary<string, VistaSite>();

                var ddrGets = new DdrGets(file: "100100.0345", iens: UserIen + ",", fields: ".02;.03",
                    optionalFlags: "I");
                rpcClient.Call(ddrGets);

                var showAllSites = "1".Equals(ddrGets.GetFieldValue(".02", false));
                var showTestSites = "1".Equals(ddrGets.GetFieldValue(".03", false));

                // empty bucket allocated and used later
                Dictionary<string, Dictionary<string, string>> allSites = null;

                const string listerFields = ".01I;.01E;.02;.03;.04;.06;100;101;102I;103I;104";

                if (showAllSites && showTestSites)
                {
                    var sitesLister = new DdrLister(file: "100100.0342", fields: listerFields);
                    rpcClient.Call(sitesLister);

                    allSites = sitesLister.GetAllRows();
                }
                else if (showAllSites && !showTestSites)
                {
                    // the screen selects all entries whose display name DOES NOT begin with the @ character
                    var screen = "if $extract($piece(^(0),U,4))'=\"@\"";
                    var ddrLister = new DdrLister(file: "100100.0342", fields: listerFields, screen: screen);
                    rpcClient.Call(ddrLister);

                    allSites = ddrLister.GetAllRows();
                }
                else if (!showAllSites && showTestSites)
                {
                    // the screen selects the sites that the user is authorized (including test sites)
                    var screen = "if $data(^R1(100100.0343,\"AU\",DUZ,+Y))";
                    var ddrLister = new DdrLister(file: "100100.0342", fields: listerFields, screen: screen);
                    rpcClient.Call(ddrLister);

                    allSites = ddrLister.GetAllRows();
                }
                else
                {
                    // the screen selects the sites that the user is authorized (NOT including test sites)
                    var screen = "if $extract($piece(^(0),U,4))'=\"@\",$data(^R1(100100.0343,\"AU\",DUZ,+Y))";
                    var ddrLister = new DdrLister(file: "100100.0342", fields: listerFields, screen: screen);
                    rpcClient.CallAndDisconnect(ddrLister); // disconnect from claims

                    allSites = ddrLister.GetAllRows();
                }

                if (allSites != null)
                {
                    // clean up work before returning the list of sites
                    foreach (var row in allSites)
                    {
                        var rowMap = row.Value;

                        var displayName = rowMap[".04"];
                        if (displayName.Contains(";"))
                            displayName = displayName.Piece(";", 1);

                        var vistaSite = new VistaSite()
                        {
                            SiteIen = rowMap[".01I"],
                            Institution = rowMap[".01"],
                            DisplayName = displayName,
                            VistaHostname = rowMap[".02"],
                            VistaPort = rowMap[".03"],
                            City = rowMap["100"],
                            State = rowMap["101"],
                            OverwriteName = rowMap["102I"].Equals("1"),
                            IgnoreSimilarName = rowMap["103I"].Equals("1"),
                            SiteCode = rowMap["104"],
                            IsTestSite = (displayName ?? "").StartsWith("@"),
                            ShellUsername = rowMap[".06"]
                        };

                        vistaSites.Add(row.Key, vistaSite);
                    }
                }

                return vistaSites;
            }
        }

        public VistaSite CanUserSyncToSite(string siteIen)
        {
            var siteIens = new List<string> {siteIen};

            var canUserSyncToSites = CanUserSyncToSites(siteIens);
            if (canUserSyncToSites.ContainsKey(siteIen))
                return canUserSyncToSites[siteIen];

            return null;
        }

        public Dictionary<string, VistaSite> CanUserSyncToSites(IEnumerable<string> siteIens)
        {
            var authorizedVistaSites = GetVistaSites();
            var syncSites = new Dictionary<string, VistaSite>();

            foreach (var siteIen in siteIens)
            {
                try
                {
                    if (authorizedVistaSites.ContainsKey(siteIen))
                        syncSites.Add(siteIen, authorizedVistaSites[siteIen]);
                }
                catch (KeyNotFoundException)
                {
                    Logger.LogError($"SECURITY: User requested access to Site:{siteIen} but is not authorized.");
                }
            }

            return syncSites;
        }
    }
}