using System.Collections.Generic;
using System.Security.Claims;
using Microsoft.Extensions.Logging;
using Vista.RPC.Procedures.DDR;
using Vista.RPC.Procedures.XOBE;
using WebVRAM.Vista.Service.Models;
using WebVRAM.Web.ng.Models;

namespace WebVRAM.Vista.Services
{
    /// <summary>
    /// VistA user service that allows a VistA user to:
    ///     1) get a vista user's details
    ///     2) determine if a user's electronic signature code is set
    ///     3) set a user's electronic signature code
    ///     4) get user's electronic signature details
    ///     5) set a user's electronic signature details
    /// </summary>
    public class VistaUserService : BaseService<VistaUserService>
    {
        public VistaUserService(ILogger<VistaUserService> logger, VistaConfig vistaConfig,
            ClaimsPrincipal claimsPrincipal) : base(logger, vistaConfig, claimsPrincipal)
        {
        }

        /// <summary>
        /// Gets a VistA user profile from the Claims VistA system
        /// </summary>
        /// <returns>a VistA user profile</returns>
        public VistaUser GetVistaUser()
        {
            using (var rpcClient = GetRpcClient())
            {
                const string ddrFields =
                    ".01;2;11;5;7;8;9;9.2;29;31.3;20.4;20.2;20.3;30;41;200.04;200.19;200.1;200.18;201";
                var ddrGets = new DdrGets(file: "200", iens: UserIen + ",", fields: ddrFields, optionalFlags: "IE");
                rpcClient.Call(ddrGets);

                var vistaUser = new VistaUser
                {
                    PersonIen = UserIen,
                    PersonName = ddrGets.GetFieldValue(".01"),

                    AccessCode = ddrGets.GetFieldValue("2", false),
                    VerifyCode = ddrGets.GetFieldValue("11", false),

                    DateOfBirth = ddrGets.GetFieldValue("5", false),
                    SocialSecurityNumber = ddrGets.GetFieldValue("9", false),
                    DisableUser = ddrGets.GetFieldValue("7"), // TODO future PUSHED "@"
                    Title = ddrGets.GetFieldValue("8"), // will not sync if doesn't exist
                    TerminationDate = ddrGets.GetFieldValue("9.2", false), // TODO future PUSHED "T+30"
                    ServiceSection = ddrGets.GetFieldValue("29"),
                    ScreenEditor = ddrGets.GetFieldValue("31.3"),
                    ElectronicSignature = ddrGets.GetFieldValue("20.4", false),
                    ElectronicSignatureTitleBlock = ddrGets.GetFieldValue("20.3"),
                    ElectronicSignaturePrintedBlock = ddrGets.GetFieldValue("20.2"),
                    CreateDate = ddrGets.GetFieldValue("30", false),
                    AllowedToUseSpooler = ddrGets.GetFieldValue("41"),
                    MultipleSignon = ddrGets.GetFieldValue("200.04"), // TODO future PUSHED "allowed"
                    MultipleSignonLimit = ddrGets.GetFieldValue("200.19"), // TODO future PUSHED "3"
                    TimedRead = ddrGets
                        .GetFieldValue("200.1"), // TODO future PUSHED 900 seconds or if number is greater then use that
                    AutoSignon = ddrGets.GetFieldValue("200.18"), // TODO futured PUSHED "no"
                    PrimaryMenu = ddrGets.GetFieldValue("201"),
                    SecondaryMenus = new Dictionary<string, string>()
                };

                // get secondary menu items
                var ddrLister = new DdrLister(file: "200.03", iens: "," + UserIen + ",", fields: ".01;.01I");
                rpcClient.Call(ddrLister);

                var allRows = ddrLister.GetAllRows();
                foreach (var row in allRows)
                {
                    var rowMap = row.Value;
                    vistaUser.SecondaryMenus.Add(row.Key, rowMap[".01"]);
                }

                // get security keys
                vistaUser.SecurityKeys = new Dictionary<string, string>();
                ddrLister = new DdrLister(file: "200.051", iens: "," + UserIen + ",", fields: ".01;.01I");
                rpcClient.Call(ddrLister);

                allRows = ddrLister.GetAllRows();
                foreach (var row in allRows)
                {
                    var rowMap = row.Value;
                    vistaUser.SecurityKeys.Add(row.Key, rowMap[".01"]);
                }

                // get cprs tabs
                vistaUser.CprsTabs = new Dictionary<string, CprsTab>();
                ddrLister = new DdrLister(file: "200.010113", iens: "," + UserIen + ",",
                    fields: ".01;.01I;.02;.03;.02I;.03I;");
                rpcClient.Call(ddrLister);

                allRows = ddrLister.GetAllRows();
                foreach (var row in allRows)
                {
                    var rowMap = row.Value;
                    var cprsTab = new CprsTab
                    {
                        CprsTabIen = row.Key,
                        CprsTabName = rowMap[".01I"],
                        EffectiveDate = rowMap[".02I"],
                        ExpirationDate = rowMap[".03I"]
                    };
                    vistaUser.CprsTabs.Add(row.Key, cprsTab);
                }

                rpcClient.Disconnect();
                
                return vistaUser;
            }
        }

        /// <summary>
        /// Determines whether or not an electronic signature code is set
        /// </summary>
        /// <returns>true if the electronic signature code is set; otherwise false</returns>
        public bool IsElectronicSignatureCodeSet()
        {
            using (var rpcClient = GetRpcClient())
            {
                var electronicSignatureDefined = new ElectronicSignatureCodeIsDefined();
                rpcClient.Call(electronicSignatureDefined);
                rpcClient.Disconnect();

                return electronicSignatureDefined.IsElectronicSignatureCodeDefined();
            }
        }

        /// <summary>
        /// Sets an electronic signature code
        /// </summary>
        /// <param name="newElectronicSignatureCode">the new electronic signature code</param>
        /// <returns>an object that represents its status</returns>
        public SetElectronicSignatureCodeResult SetElectronicSignatureCode(string newElectronicSignatureCode)
        {
            using (var rpcClient = GetRpcClient())
            {
                var electronicSignatureSetCode = new ElectronicSignatureSetCode(newElectronicSignatureCode);
                rpcClient.Call(electronicSignatureSetCode);
                rpcClient.Disconnect();

                if (electronicSignatureSetCode.IsElectronicSignatureSet())
                {
                    return new SetElectronicSignatureCodeResult() {ElectronicSignatureChanged = true};
                }

                return new SetElectronicSignatureCodeResult()
                {
                    ElectronicSignatureChanged = false,
                    ErrorMessage = electronicSignatureSetCode.GetMessage()
                };
            }
        }

        /// <summary>
        /// Get the user's electronic signature details 
        /// </summary>
        /// <returns>an object that represents the electronic signature details</returns>
        public ElectronicSignatureDetails GetElectronicSignatureDetails()
        {
            using (var rpcClient = GetRpcClient())
            {
                var electronicSignatureGetDetails = new ElectronicSignatureGetDetails();
                rpcClient.Call(electronicSignatureGetDetails);
                rpcClient.Disconnect();

                return new ElectronicSignatureDetails()
                {
                    DigitalPagerNumber = electronicSignatureGetDetails.DigitalPagerNumber,
                    Initials = electronicSignatureGetDetails.Initials,
                    OfficePhoneNumber = electronicSignatureGetDetails.OfficePhoneNumber,
                    SignatureBlockPrintedName = electronicSignatureGetDetails.SignatureBlockPrintedName,
                    SignatureBlockTitle = electronicSignatureGetDetails.SignatureBlockTitle,
                    VoicePagerNumber = electronicSignatureGetDetails.VoicePagerNumber
                };
            }
        }

        /// <summary>
        /// Sets a user's electronic signature details
        /// </summary>
        /// <param name="details">the user's updated electronic signature details</param>
        /// <returns>an object that represents its status</returns>
        public SetElectronicSignatureDetailsResult SetElectronicSignatureDetails(ElectronicSignatureDetails details)
        {
            using (var rpcClient = GetRpcClient())
            {
                var electronicSignatureSetDetails = new ElectronicSignatureSetDetails(initials: details.Initials,
                    signatureBlockPrintedName: details.SignatureBlockPrintedName,
                    signatureBlockTitle: details.SignatureBlockTitle, officePhoneNumber: details.OfficePhoneNumber,
                    voicePagerNumber: details.VoicePagerNumber, digitalPagerNumber: details.VoicePagerNumber);

                rpcClient.Call(electronicSignatureSetDetails);
                rpcClient.Disconnect();

                if (electronicSignatureSetDetails.IsSetDetailsSuccessful())
                {
                    return new SetElectronicSignatureDetailsResult()
                    {
                        ElectronicSignatureDetailsChanged = electronicSignatureSetDetails.IsSetDetailsSuccessful(),
                    };
                }

                return new SetElectronicSignatureDetailsResult()
                {
                    ErrorMessage = electronicSignatureSetDetails.GetMessage()
                };
            }
        }
    }
}