﻿using System;
using System.Collections.Generic;
using System.IdentityModel.Tokens.Jwt;
using System.Security;
using System.Security.Claims;
using System.Text;
using Microsoft.Extensions.Logging;
using Microsoft.IdentityModel.Tokens;
using Vista.RPC;
using WebVRAM.Web.ng.Models;

namespace WebVRAM.Web.ng.Common
{
    public sealed class JwtUtility
    {
        private ILogger<JwtUtility> _logger;
        private JwtConfig _jwtConfig;
        
        public JwtUtility(ILogger<JwtUtility> logger, JwtConfig jwtConfig)
        {
            logger.LogDebug("JwtUtility instantiated");

            _logger = logger ?? throw new ArgumentNullException(nameof(logger));
            _jwtConfig = jwtConfig ?? throw new ArgumentNullException(nameof(jwtConfig));

            try
            {
                JwtExpiration = jwtConfig.JwtExpiration;
                JwtSecurityKey = jwtConfig.JwtSecurityKey;
                JwtSignatureKey = jwtConfig.JwtSignatureKey;
                JwtRegisteredClaims = jwtConfig.JwtRegisteredClaims;

                _symmetricSecurityKey = new SymmetricSecurityKey(Encoding.ASCII.GetBytes(JwtSecurityKey.ToClearText()));
                _symmetricSignatureKey = new SymmetricSecurityKey(Encoding.ASCII.GetBytes(JwtSignatureKey.ToClearText()));

                _signingCredentials = new SigningCredentials(_symmetricSignatureKey,
                    SecurityAlgorithms.HmacSha512Signature, SecurityAlgorithms.HmacSha512);

                _encryptingCredentials = new EncryptingCredentials(_symmetricSecurityKey, SecurityAlgorithms.Aes256KW,
                    SecurityAlgorithms.Aes256CbcHmacSha512);

                _jwtSecurityTokenHandler = new JwtSecurityTokenHandler();
            }
            catch (FormatException e)
            {
                _logger.LogError(e,"failed to create jwt security handler because of an improperly formatted parse");
            }
            catch (OverflowException e)
            {
                _logger.LogError(e,"failed to create jwt security handler because of an integer overflow when parsing");
            }
            catch (ArgumentNullException e)
            {
                _logger.LogError(e,$"failed to create jwt security handler because a credential argument was null: {e.ParamName}");
            }
        }

        private SigningCredentials _signingCredentials;
        private EncryptingCredentials _encryptingCredentials;
        private JwtSecurityTokenHandler _jwtSecurityTokenHandler;
        private SymmetricSecurityKey _symmetricSecurityKey;
        private SymmetricSecurityKey _symmetricSignatureKey;

        // default values
        private int JwtExpiration = 7;
        private SecureString JwtSignatureKey;
        private SecureString JwtSecurityKey;
        private string JwtRegisteredClaims = string.Empty;

        public double GetJwtEpoch()
        {
            var originDate = new DateTime(1970, 1, 1, 0, 0, 0, 0, DateTimeKind.Utc);
            var timeSpan = DateTime.Now.ToUniversalTime() - originDate;
            return Math.Floor(timeSpan.TotalSeconds);
        }

        public double GetJwtDateTime(DateTime dateTime)
        {
            var originDate = new DateTime(1970, 1, 1, 0, 0, 0, 0, DateTimeKind.Utc);
            var timeSpan = dateTime.ToUniversalTime() - originDate;
            return Math.Floor(timeSpan.TotalSeconds);
        }

        /// <summary>
        /// Get a JWT token built using the JWT specified in the appsettings
        /// </summary>
        /// <param name="claims"></param>
        /// <returns></returns>
        public string GetJwtToken(IList<Claim> claims)
        {
            try
            {
                var tokenHandler = _jwtSecurityTokenHandler;

                var tokenDescriptor = new SecurityTokenDescriptor
                {
                    Subject = new ClaimsIdentity(claims),
                    Expires = DateTime.UtcNow.AddDays(JwtExpiration),
                    SigningCredentials = _signingCredentials,
                    EncryptingCredentials = _encryptingCredentials
                };

                var token = tokenHandler.CreateToken(tokenDescriptor);
                var tokenString = tokenHandler.WriteToken(token);

                return tokenString;
            }
            catch (ArgumentNullException e)
            {
                _logger.LogError(e,$"failed to write jwt; argument null {e.ParamName}");
            }
            catch (ArgumentException e)
            {
                _logger.LogError(e,$"failed to write jwt; argument invalid ${e.ParamName} - ${e.Message}");
            }
            catch (SecurityTokenEncryptionFailedException e)
            {
                _logger.LogError(e,"failed to encrypt");
            }

            return null;
        }

    }
}