﻿using System;
using System.Collections.Generic;
using System.Security.Claims;
using Microsoft.AspNetCore.Authorization;
using Microsoft.AspNetCore.Http;
using Microsoft.AspNetCore.Mvc;
using Microsoft.Extensions.Logging;
using Vista.RPC;
using WebVRAM.Vista.Services;
using WebVRAM.Web.ng.Common;
using WebVRAM.Web.ng.Controllers.MVC;
using WebVRAM.Web.ng.Models;

namespace WebVRAM.Web.ng.Controllers
{
    [AllowAnonymous]
    public class LoginController : BaseApiController<LoginController>
    {
        private LoginService _loginService;
        private JwtUtility _jwtUtility;

        public LoginController(ILogger<LoginController> logger, LoginService loginService, JwtUtility jwtUtility) : base(logger)
        {
            _loginService = loginService;
            _jwtUtility = jwtUtility;
        }

        private static string _claimsWelcomeMessage = null;

        [HttpGet]
        public IActionResult GetClaimsWelcomeMessage()
        {
            _logger.LogDebug(nameof(GetClaimsWelcomeMessage));

            if (_claimsWelcomeMessage == null)
            {
                _claimsWelcomeMessage = _loginService.GetClaimsWelcomeMessage();
                _logger.LogDebug("fetching claims welcome message from the vista server");
            }

            return Ok(_claimsWelcomeMessage);
        }

        [HttpPost]
        public IActionResult Login([FromBody] LoginInput loginInput)
        {
            var tryGetValue = HttpContext.Request.Cookies.TryGetValue(nameof(HomeController.WEBVRAM), out var expectedXsrf);
            if (!tryGetValue || loginInput.Xsrf != expectedXsrf)
            {
                var remoteIpAddress = HttpContext.Connection.RemoteIpAddress.ToString();
                _logger.LogWarning($"Security Event: Someone tried to login to WebVRAM from outside the application. Remote IP: {remoteIpAddress}");

                return Unauthorized();
            }

            if (loginInput.ChangeVerifyCode && !string.IsNullOrEmpty(loginInput.NewVerifyCode))
            {
                var changeVerifyCodeResult = _loginService.ChangeVerifyCode(loginInput.AccessCode.ToSecureString(),
                    loginInput.VerifyCode.ToSecureString(), loginInput.NewVerifyCode.ToSecureString());
                if (changeVerifyCodeResult.VerifyCodeChanged)
                {
                    loginInput.VerifyCode = loginInput.NewVerifyCode;
                }
            }

            var loginResult = _loginService.Login(loginInput.AccessCode.ToSecureString(),
                loginInput.VerifyCode.ToSecureString());

            if (!loginResult.LoggedIn)
            {
                return Unauthorized();
            }

            if (loginResult.VerifyNeedsChanging)
            {
                return Ok(new LoginResponse()
                {
                    VerifyCodeNeedsChanging = true
                });
            }

            var claims = new List<Claim>()
            {
                new Claim(ClaimTypes.Name, loginResult.PersonName),
                new Claim(JwtPrivateClaims.Xsrf, expectedXsrf),
                new Claim(JwtPrivateClaims.AccessCode, loginInput.AccessCode),
                new Claim(JwtPrivateClaims.VerifyCode, loginInput.VerifyCode),
                new Claim(JwtPrivateClaims.ClaimsDUZ, loginResult.LoggedInDuz),
                new Claim(JwtPrivateClaims.BseToken, loginResult.BseToken)
            };

            var jwtToken = _jwtUtility.GetJwtToken(claims);
            var loginOutput = new LoginResponse()
            {
                JwtToken = jwtToken,
                PersonName = loginResult.PersonName
            };
            
            return Ok(loginOutput);
        }
    }
}