﻿using System;
using System.Security.Claims;
using System.Security.Principal;
using System.Text;
using System.Threading.Tasks;
using Microsoft.AspNetCore.Authentication.JwtBearer;
using Microsoft.AspNetCore.Authorization;
using Microsoft.AspNetCore.Builder;
using Microsoft.AspNetCore.Hosting;
using Microsoft.AspNetCore.Http;
using Microsoft.AspNetCore.Mvc;
using Microsoft.AspNetCore.Server.IISIntegration;
using Microsoft.Extensions.Configuration;
using Microsoft.Extensions.DependencyInjection;
using Microsoft.Extensions.Logging;
using Microsoft.IdentityModel.Tokens;
using Vista.RPC;
using WebVRAM.Data.Repository;
using WebVRAM.Vista.Service;
using WebVRAM.Vista.Services;
using WebVRAM.Web.ng.Common;
using WebVRAM.Web.ng.Middleware;
using WebVRAM.Web.ng.Models;
using WebVRAM.Web.ng.SignalR;

namespace WebVRAM.Web.ng
{
    public class Startup
    {
        public Startup(ILogger<Startup> logger, IConfiguration configuration)
        {
            _logger = logger ?? throw new ArgumentNullException(nameof(logger));
            Configuration = configuration ?? throw new ArgumentNullException(nameof(configuration));
        }

        private ILogger _logger;

        public IConfiguration Configuration { get; }

        // This method gets called by the runtime. Use this method to add services to the container.
        public void ConfigureServices(IServiceCollection services)
        {
            var appConfigSection = Configuration.GetSection("WebVRAM");
            var vistaConfigSection = appConfigSection.GetSection("Vista");
            var jwtConfigSection = appConfigSection.GetSection("JWT");

            services.AddAuthentication(options =>
            {
                options.DefaultAuthenticateScheme = IISDefaults.AuthenticationScheme;
                options.DefaultChallengeScheme = JwtBearerDefaults.AuthenticationScheme;
            }).AddJwtBearer(JwtBearerDefaults.AuthenticationScheme, options =>
            {
#if DEBUG
                options.RequireHttpsMetadata = false;
#endif
                var symmetricSignatureKey =
                new SymmetricSecurityKey(Encoding.ASCII.GetBytes(jwtConfigSection["JwtSignatureKey"]));
                var symmetricSecurityKey =
                    new SymmetricSecurityKey(Encoding.ASCII.GetBytes(jwtConfigSection["JwtSecurityKey"]));

                options.TokenValidationParameters = new TokenValidationParameters()
                {
                    ValidateAudience = false,
                    ValidateIssuer = false,
                    ValidateActor = false,

                    RequireExpirationTime = true,
                    ClockSkew = TimeSpan.FromMinutes(5),

                    RequireSignedTokens = true,
                    ValidateIssuerSigningKey = true,
                    IssuerSigningKey = symmetricSignatureKey,
                    TokenDecryptionKey = symmetricSecurityKey
                };

                /*
                 In standard web APIs, bearer tokens are sent in an HTTP header. However, SignalR
                 is unable to set these headers in browsers when using some transports. When using 
                 WebSockets and Server-Sent Events, the token is transmitted as a query string 
                 parameter. In order to support this on the server, additional configuration is required.
                 https://docs.microsoft.com/en-us/aspnet/core/signalr/authn-and-authz?view=aspnetcore-2.2
                 */
                options.Events = new JwtBearerEvents()
                {
                    OnMessageReceived = context =>
                    {
                        var accessToken = context.Request.Query["access_token"];

                        // If the request is for our hub...
                        var path = context.HttpContext.Request.Path;
                        if (!string.IsNullOrEmpty(accessToken) &&
                            (path.StartsWithSegments("/vistaHub")))
                        {
                            // Read the token out of the query string
                            context.Token = accessToken;
                        }

                        return Task.CompletedTask;
                    }
                };
            });

            services.AddAuthorization(config =>
            {
                var authorizationPolicy = new AuthorizationPolicyBuilder()
                    .AddAuthenticationSchemes(JwtBearerDefaults.AuthenticationScheme).RequireClaim(ClaimTypes.Name);

                config.AddPolicy(JwtBearerDefaults.AuthenticationScheme, authorizationPolicy.Build());
            });

            services.AddMemoryCache();
            services.AddMvc().SetCompatibilityVersion(CompatibilityVersion.Version_2_2);
            
            services.AddSignalR();
            
            // configure dependency injection
            services.AddHttpContextAccessor();

            try
            {
                var vistaConfig = BuildVistaConfig(vistaConfigSection);
                services.AddSingleton<VistaConfig>(provider => vistaConfig);

                var jwtConfig = BuildJwtConfig(jwtConfigSection);
                services.AddSingleton<JwtConfig>(provider => jwtConfig);
            }
            catch (ApplicationException e)
            {
                _logger.LogCritical("--------------------------------------------------------");
                _logger.LogCritical("Failed to read configuration for application start up...");
                _logger.LogCritical("--------------------------------------------------------");
                _logger.LogCritical(e, e.Message);
                Environment.Exit(-1);
            }

            services.AddTransient<LoginService>();
            services.AddTransient<VistaUserService>();
            services.AddTransient<SyncService>();
            services.AddTransient<JwtUtility>();

            services.AddScoped<ClaimsPrincipal>(
                provider => provider.GetService<IHttpContextAccessor>()?.HttpContext?.User);

            /*
            services.AddScoped<UnitOfWork>(provider =>
            {
                var entitiesFactory = new WebVRAM_EntitiesFactory("");
                return new UnitOfWork(entitiesFactory);
            });
            */
        }

        // This method gets called by the runtime. Use this method to configure the HTTP request pipeline.
        public void Configure(IApplicationBuilder app, IHostingEnvironment env)
        {
            if (env.IsDevelopment())
            {
                _logger.LogDebug("-------------------------------");
                _logger.LogDebug("| WebVRAM Application Startup |");
                _logger.LogDebug("-------------------------------");
                _logger.LogDebug("| ASP.NET Core IsDevelopment  |");
                _logger.LogDebug("-------------------------------");
                app.UseDeveloperExceptionPage();
            }
            else
            {
                app.UseMiddleware<ExceptionMiddleware>();
                app.UseHsts();
            }

            _logger.LogDebug($"Running in Environment: {env.EnvironmentName}");
            
            app.UseAuthentication();

#if RELEASE
            app.UseHttpsRedirection();
#endif

            app.UseStaticFiles();

            app.UseSignalR(routes => { routes.MapHub<VistaHub>("/vistaHub"); });

            app.UseMvc(routes =>
            {
                routes.MapRoute(
                    name: "default",
                    template: "{controller=Home}/{action=Index}");
            });
        }

        private JwtConfig BuildJwtConfig(IConfigurationSection section)
        {
            var jwtRegisteredClaims = section["JwtRegisteredClaims"];

            if (string.IsNullOrEmpty(jwtRegisteredClaims))
            {
                var errorMessage = "failed to parse the JwtRegisteredClaims in the appsettings.json";

                _logger.LogCritical(errorMessage);
                throw new ApplicationException(errorMessage);
            }

            var stringJwtExpiration = section["JwtExpiration"];
            var tryParse = int.TryParse(stringJwtExpiration, out var jwtExpiration);
            if (!tryParse)
            {
                var errorMessage = "failed to parse the JwtExpiration in the appsettings.json";

                _logger.LogCritical(errorMessage);
                throw new ApplicationException(errorMessage);
            }

            var jwtSecurityKeyClearText = section["JwtSecurityKey"];
            if (string.IsNullOrEmpty(jwtSecurityKeyClearText))
            {
                var errorMessage = "failed to parse the JwtSecurityKey in the appsettings.json";

                _logger.LogCritical(errorMessage);
                throw new ApplicationException(errorMessage);
            }

            if (jwtSecurityKeyClearText.Length != 32)
            {
                var errorMessage = "JwtSecurityKey in appsettings.json must be 32 bytes";

                _logger.LogCritical(errorMessage);
                throw new ApplicationException(errorMessage);
            }

            var jwtSecurityKey = jwtSecurityKeyClearText.ToSecureString();

            var jwtSignatureKeyClearText = section["JwtSignatureKey"];
            if (string.IsNullOrEmpty(jwtSignatureKeyClearText))
            {
                var errorMessage = "failed to parse the JwtSignatureKey in appsettings.json";
                _logger.LogCritical(errorMessage);
                throw new ApplicationException(errorMessage);
            }

            if (jwtSignatureKeyClearText.Length < 64)
            {
                var errorMessage = "JwtSignatureKey in appsettings.json must be >= 64 bytes";
                _logger.LogCritical(errorMessage);
                throw new ApplicationException(errorMessage);
            }

            var jwtSignatureKey = jwtSignatureKeyClearText.ToSecureString();

            return new JwtConfig(jwtRegisteredClaims, jwtExpiration, jwtSecurityKey, jwtSignatureKey);
        }

        private VistaConfig BuildVistaConfig(IConfigurationSection section)
        {
            var vistaMenuOptionClearText = section["MenuOption"];
            if (string.IsNullOrEmpty(vistaMenuOptionClearText))
            {
                var errorMessage = "failed to parse the Hostname in the appsettings.json";

                _logger.LogCritical(errorMessage);
                throw new ApplicationException(errorMessage);
            }

            var vistaMenuOption = vistaMenuOptionClearText.ToSecureString();

            var bseApplicationPassPhraseClearText = section["BseApplicationPassPhrase"];
            if (string.IsNullOrEmpty(bseApplicationPassPhraseClearText))
            {
                var errorMessage = "failed to parse the BseApplicationPassPhrase in the appsettings.json";

                _logger.LogCritical(errorMessage);
                throw new ApplicationException(errorMessage);
            }

            var bseApplicationPassPhrase = bseApplicationPassPhraseClearText.ToSecureString();
            
            var hostname = section["Hostname"];
            if (string.IsNullOrEmpty(hostname))
            {
                var errorMessage = "failed to parse the Hostname in the appsettings.json";
                
                _logger.LogCritical(errorMessage);
                throw new ApplicationException(errorMessage);
            }

            var stringRpcPort = section["RpcPort"];
            var tryParse = int.TryParse(stringRpcPort, out var rpcPort);
            if (!tryParse)
            {
                var errorMessage = "failed to parse the RpcPort in the appsettings.json";

                _logger.LogCritical(errorMessage);
                throw new ApplicationException(errorMessage);
            }

            return new VistaConfig(claimsHostname: hostname, claimsPort: rpcPort, vistaMenuOption: vistaMenuOption,
                vistaBsePassPhrase: bseApplicationPassPhrase);
        }
    }
}